<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2023 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\Component\ATS\Administrator\CliCommand;

use Akeeba\Component\ATS\Administrator\CliCommand\MixIt\ConfigureIO;
use Akeeba\Component\ATS\Administrator\CliCommand\MixIt\IsPro;
use Akeeba\Component\ATS\Administrator\CliCommand\MixIt\MemoryInfo;
use Akeeba\Component\ATS\Administrator\CliCommand\MixIt\TimeInfo;
use Akeeba\Component\ATS\Administrator\Mixin\ComponentObjectTrait;
use Exception;
use Joomla\CMS\Component\ComponentHelper;
use Joomla\CMS\Language\Text;
use Joomla\CMS\MVC\Factory\MVCFactoryAwareTrait;
use Joomla\Database\DatabaseAwareTrait;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Output\OutputInterface;

defined('_JEXEC') or die;

class Config extends \Joomla\Console\Command\AbstractCommand
{
	use ConfigureIO;
	use IsPro;
	use MemoryInfo;
	use TimeInfo;
	use MVCFactoryAwareTrait;
	use DatabaseAwareTrait;
	use ComponentObjectTrait;

	/**
	 * The default command name
	 *
	 * @var    string
	 * @since  5.0.0
	 */
	protected static $defaultName = 'ats:config';

	/**
	 * @inheritDoc
	 */
	protected function doExecute(InputInterface $input, OutputInterface $output): int
	{
		// Configure the Symfony I/O
		$this->configureSymfonyIO($input, $output);

		$key     = $input->getOption('key') ?? null;
		$unset   = $input->getOption('unset') ?? false;
		$value   = $unset ? null : $input->getOption('value') ?? null;
		$machine = $input->getOption('machine') ?? false;

		$params = ComponentHelper::getParams('com_ats');

		if (empty($key) && $machine)
		{
			echo $params->toString('JSON');

			return 0;
		}

		if (empty($key))
		{
			$asArray = $params->toArray();
			$asArray = array_map(function ($v) {
				return is_scalar($v) ? $v : json_encode($v);
			}, $asArray);
			$this->ioStyle->table(['Key', 'Value'], array_map(function ($k, $v) {
				return [$k, $v];
			}, array_keys($asArray), array_values($asArray)));

			return 0;
		}

		if (is_null($value) && !$unset && $machine)
		{
			echo json_encode($params->get($key));

			return 0;
		}

		if (is_null($value) && !$unset)
		{
			$this->ioStyle->table(['Key', 'Value'], [
				[$key, $params->get($key)]
			]);

			return 0;
		}

		if (!is_null($value))
		{
			if (
				is_string($value)
				&& (strlen($value) > 14)
				&& substr($value, 0, 1) === '{'
				&& substr($value, -1) === '}') {
				try
				{
					$json_decoded = @json_decode($value, true);
				}
				catch (Exception $e)
				{
					$json_decoded = null;
				}

				$value = is_array($json_decoded) ? $json_decoded : $value;
			}

			$params->set($key, $value);
		}
		elseif ($unset)
		{
			$params->remove($key);
		}

		try
		{
			$this->getExtensionObject()
				->getComponentParametersService()
				->save($params);
		}
		catch (Exception $e)
		{
			if ($machine)
			{
				echo json_encode([
					'code'  => $e->getCode(),
					'error' => $e->getMessage(),
				]);
			}
			else
			{
				$this->ioStyle->error([
					sprintf("[%s] %s", $e->getCode(), $e->getMessage()),
					sprintf("%s::L%u", $e->getFile(), $e->getLine()),
					$e->getTraceAsString()
				]);
			}

			return 1;
		}

		return 0;
	}

	/**
	 * Configure the command.
	 *
	 * @return  void
	 *
	 * @since   7.5.0
	 */
	protected function configure(): void
	{
		$this->setDescription(Text::_('COM_ATS_CLI_CONFIG_DESC'));
		$this->setHelp(Text::_('COM_ATS_CLI_CONFIG_HELP'));

		$this->addOption('key', 'k', InputOption::VALUE_REQUIRED, Text::_('COM_ATS_CLI_CONFIG_KEY'));
		$this->addOption('value', 'a', InputOption::VALUE_OPTIONAL, Text::_('COM_ATS_CLI_CONFIG_VALUE'));
		$this->addOption('unset', 'u', InputOption::VALUE_NONE, Text::_('COM_ATS_CLI_CONFIG_UNSET'));
		$this->addOption('machine', 'm', InputOption::VALUE_NONE, Text::_('COM_ATS_CLI_CONFIG_MACHINE'));
	}

}