<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2023 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\Component\ATS\Administrator\CliCommand;

defined('_JEXEC') or die;

use Akeeba\Component\ATS\Administrator\CliCommand\MixIt\CliRouting;
use Akeeba\Component\ATS\Administrator\CliCommand\MixIt\ConfigureIO;
use Akeeba\Component\ATS\Administrator\CliCommand\MixIt\IsPro;
use Akeeba\Component\ATS\Administrator\CliCommand\MixIt\MemoryInfo;
use Akeeba\Component\ATS\Administrator\CliCommand\MixIt\TimeInfo;
use Akeeba\Component\ATS\Administrator\Exceptions\CronCommandMissing;
use Akeeba\Component\ATS\Administrator\Exceptions\CronCommandNotFound;
use Akeeba\Component\ATS\Administrator\Helper\Cron as CronHelper;
use Akeeba\Component\ATS\Administrator\Helper\Logger;
use Joomla\CMS\Factory;
use Joomla\CMS\Language\Text;
use Joomla\CMS\Log\Log;
use Joomla\CMS\Log\LogEntry;
use Joomla\CMS\MVC\Factory\MVCFactoryAwareTrait;
use Joomla\CMS\Router\Route;
use Joomla\Console\Command\AbstractCommand;
use Joomla\Database\DatabaseAwareTrait;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Output\OutputInterface;

class Cron extends AbstractCommand
{
	use ConfigureIO;
	use IsPro;
	use MemoryInfo;
	use TimeInfo;
	use CliRouting;
	use MVCFactoryAwareTrait;
	use DatabaseAwareTrait;

	/**
	 * The default command name
	 *
	 * @var    string
	 * @since  5.0.0
	 */
	protected static $defaultName = 'ats:cron';

	/**
	 * @inheritDoc
	 */
	protected function doExecute(InputInterface $input, OutputInterface $output): int
	{
		// Configure the Symfony I/O
		$this->configureSymfonyIO($input, $output);

		// Disable database query logging (causes out–of–memory errors)
		$this->setDatabase(Factory::getContainer()->get('DatabaseDriver'));
		$this->getDatabase()->setMonitor(null);

		// Initialise the CLI routing
		$this->initCliRouting();

		/**
		 * Register the HTML helper.
		 *
		 * This goes a bit sideways to get there for a good reason.
		 *
		 * Calling the CMS Router boots the component which registers the HTML helper. If we don't do that and try to
		 * register the HTML helper directly the next time we try to build a URL through the Router it will try to boot
		 * the component which will try to re-register the HTML helper. Since the helper is already registered it will
		 * bomb out.
		 */
		Route::_('index.php?option=com_ats');

		// Get the CRON task to execute
		$task = $input->getArgument('task');
		$verbose = $input->getOption('verbose');
		$debug = $input->getOption('debug');

		if ($debug)
		{
			Factory::getApplication()->set('debug', 1);
		}

		if (!defined('JDEBUG'))
		{
			define('JDEBUG', $debug ? 1 : 0);
		}

		$this->ioStyle->title(Text::sprintf('COM_ATS_CLI_CRON_HEAD', $task));

		$this->ioStyle->comment([
			'Akeeba Ticket System',
			sprintf('Copyright (c) 2011-%s Akeeba Ltd / Nicholas K. Dionysopoulos', gmdate('Y')),
			'Akeeba Ticket System is Free Software, distributed under the terms of the GNU General Public License version 3 or, at your option, any later version. This program comes with ABSOLUTELY NO WARRANTY as per sections 15 & 16 of the license. See http://www.gnu.org/licenses/gpl-3.0.html for details.',
			sprintf('PHP %s (%s)', PHP_VERSION, PHP_SAPI),
		]);

		/**
		 * Output the log entries to the output if verbose mode is enabled
		 *
		 * @param   LogEntry  $entry  The log entry to process
		 *
		 * @return  void
		 * @since   5.0.0
		 */
		$logCallback = function (LogEntry $entry) use ($verbose) {
			if (!$verbose)
			{
				return;
			}

			switch ($entry->priority)
			{
				case Log::EMERGENCY:
				case Log::ALERT:
				case Log::CRITICAL:
				case Log::ERROR:
					$this->ioStyle->error($entry->message);
					break;

				case Log::WARNING:
					$this->ioStyle->warning($entry->message);
					break;

				case Log::NOTICE:
					$this->ioStyle->comment($entry->message);
					break;

				case Log::INFO:
					$this->ioStyle->info($entry->message);
					break;

				case Log::DEBUG:
				default:
					$this->ioStyle->text($entry->message);
					break;
			}
		};

		if ($verbose)
		{
			Logger::registerCLIOutputLogger('ats.cron', $logCallback);
			Logger::registerCLIOutputLogger('ats.emails', $logCallback);

			$this->ioStyle->info(Text::sprintf('COM_ATS_CLI_CRON_LBL_CURRENT_MEM_USAGE', $this->memUsage()));
		}

		$startTime = microtime(true);

		CronHelper::log('Starting the CRON job from CLI');
		CronHelper::log('Executing the command');

		try
		{
			$return = CronHelper::run($task);

			CronHelper::log('Exit: CRON execution finished');
		}
		catch (CronCommandMissing $e)
		{
			CronHelper::log('There was no command specified in the command line', Log::ERROR);
			CronHelper::log('Exit: no command');

			$this->ioStyle->error($e->getMessage());

			return 1;
		}
		catch (CronCommandNotFound $e)
		{
			CronHelper::log(sprintf('The requested command, “%s”, is not implemented', $task), Log::ERROR);
			CronHelper::log('Exit: unknown command');

			$this->ioStyle->error($e->getMessage());

			return 2;
		}
		catch (\Exception $e)
		{
			CronHelper::log(sprintf('An error occurred executing command “%s”', $task), Log::ERROR);
			CronHelper::log($e->getMessage());
			CronHelper::log($e->getFile() . '::' . $e->getLine());

			$this->ioStyle->error(array_merge([$e->getMessage()], explode("\n", $e->getTraceAsString())));

			CronHelper::log('Exit: error executing command');

			return 255;
		}

		if ($verbose)
		{
			$this->ioStyle->info([
				Text::sprintf('COM_ATS_CLI_CRON_LBL_CURRENT_MEM_USAGE', $this->memUsage()),
				Text::sprintf('COM_ATS_CLI_CRON_LBL_PEAK_MEM_USAGE', $this->peakMemUsage()),
			]);
		}

		if (!$return)
		{
			$this->ioStyle->error(
				Text::sprintf('COM_ATS_CLI_CRON_LBL_DONE_FAILED', $this->timeAgo((int) $startTime, time(), '', false))
			);

			return 127;
		}

		$this->ioStyle->success(
			Text::sprintf('COM_ATS_CLI_CRON_LBL_DONE_SUCCESS', $this->timeAgo((int) $startTime, time(), '', false))
		);

		return 0;
	}

	/**
	 * Configure the command.
	 *
	 * @return  void
	 *
	 * @since   7.5.0
	 */
	protected function configure(): void
	{
		$this->setDescription(Text::_('COM_ATS_CLI_CRON_DESC'));
		$this->setHelp(Text::_('COM_ATS_CLI_CRON_HELP'));

		$this->addArgument('task', InputOption::VALUE_REQUIRED, Text::_('COM_ATS_CLI_CRON_TASK'));
		$this->addOption('debug', 'd', InputOption::VALUE_NONE | InputOption::VALUE_OPTIONAL, Text::_('COM_ATS_CLI_CRON_DEBUG'));
	}

}