<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2023 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\Component\ATS\Administrator\Controller;

defined('_JEXEC') or die;

use Akeeba\Component\ATS\Administrator\Helper\Permissions;
use Akeeba\Component\ATS\Administrator\Mixin\ControllerEventsTrait;
use Akeeba\Component\ATS\Administrator\Mixin\ControllerReturnURLTrait;
use Akeeba\Component\ATS\Administrator\Mixin\ControllerReusableModelsTrait;
use Akeeba\Component\ATS\Administrator\Table\AttachmentTable;
use Exception;
use Joomla\CMS\Factory;
use Joomla\CMS\Filesystem\File;
use Joomla\CMS\Filesystem\Path;
use Joomla\CMS\Language\Text;
use Joomla\CMS\MVC\Controller\AdminController;
use Joomla\Utilities\ArrayHelper;
use RuntimeException;

class AttachmentsController extends AdminController
{
	use ControllerEventsTrait;
	use ControllerReusableModelsTrait;
	use ControllerReturnURLTrait;

	/** @inheritdoc */
	protected $text_prefix = 'COM_ATS_ATTACHMENTS';

	/**
	 * Downloads an attachment.
	 *
	 * @throws Exception
	 * @since  5.0.0
	 */
	public function download()
	{
		// Check the CSRF token, prevents hotlinking of attachments
		$this->checkToken('get');

		// Make sure there is a valid attachment ID
		$cid = $this->input->getInt('cid') ?: [];
		$cid = is_array($cid) ? $cid : [$cid];
		$cid = ArrayHelper::toInteger($cid);

		if (empty($cid))
		{
			throw new RuntimeException(Text::_('JERROR_ALERTNOAUTHOR'), 403);
		}

		$id = array_pop($cid);

		if (empty($id))
		{
			throw new RuntimeException(Text::_('JERROR_ALERTNOAUTHOR'), 403);
		}

		// Get the attachment record
		/** @var AttachmentTable $attachment */
		$attachment = $this->getModel()->getTable('Attachment', 'Administrator');

		if (!$attachment->load($id))
		{
			throw new RuntimeException(Text::_('JERROR_ALERTNOAUTHOR'), 403);
		}

		// Make sure the user is allowed to view the attachment
		if (!Permissions::attachmentVisible($attachment))
		{
			throw new RuntimeException(Text::_('JERROR_ALERTNOAUTHOR'), 403);
		}

		// Pretend this is a raw view
		$this->app->input->set('format', 'raw');

		// Disable error reporting and error display
		if (function_exists('error_reporting'))
		{
			error_reporting(0);
		}

		if (function_exists('ini_set'))
		{
			@ini_set('display_error', 0);
		}

		// Clear the output buffer cache
		while (@ob_end_clean())
		{
		}

		// Calculate the Etag
		$etagContent = implode('', array_values($attachment->getProperties()));
		$eTag        = function_exists('sha1') ? sha1($etagContent) : md5($etagContent);

		// Do we have an If-None-Match header?
		$ifNoneMatchDefault = null;

		if (function_exists('apache_request_headers'))
		{
			$headers = apache_request_headers();

			if (array_key_exists('If-None-Match', $headers))
			{
				$ifNoneMatchDefault = $headers['If-None-Match'];
			}
		}

		$ifNoneMatch = $this->input->server->getRaw('HTTP_IF_NONE_MATCH', $ifNoneMatchDefault);

		if ($ifNoneMatch == $eTag)
		{
			header('HTTP/1.0 304 Not Modified');

			$this->app->close();
		}

		$filePath = $attachment->getAbsoluteFilename();
		$filepath = Path::clean($filePath);
		$basename = $attachment->original_filename;

		if (!File::exists($filepath))
		{
			header('HTTP/1.0 404 Not Found');

			$this->app->close();
		}

		$header_file = $basename;

		// Fix IE bugs
		$userAgent = $this->input->server->get('HTTP_USER_AGENT', null);

		if (is_string($userAgent) && strstr($userAgent, 'MSIE'))
		{
			$header_file = preg_replace('/\./', '%2e', $basename, substr_count($basename, '.') - 1);

			if (ini_get('zlib.output_compression'))
			{
				ini_set('zlib.output_compression', 'Off');
			}
		}

		@clearstatcache();

		// Send the relevant headers
		$date     = Factory::getDate($attachment->created);
		$filesize = @filesize($filepath);

		header('Date: ' . $date->toRFC822());
		header('Etag: ' . $eTag);
		header("Content-Description: File Transfer");
		header('Content-Type: ' . ($attachment->mime_type ?: 'application/octet-stream'));
		header("Accept-Ranges: bytes");
		header(sprintf("Content-Disposition: attachment; filename=\"%s\"", $header_file));
		header('Content-Transfer-Encoding: binary');

		if ($filesize > 0)
		{
			header(sprintf("Content-Length: %d", (int) $filesize));
		}

		// Disable time limits
		@set_time_limit(0);

		// Use 1M chunks for echoing the data to the browser
		@flush();
		$chunksize = 1048576;

		$handle = @fopen($filepath, 'r');

		if ($handle !== false)
		{
			while (!feof($handle))
			{
				$buffer = fread($handle, $chunksize);

				echo $buffer;

				@ob_flush();
				flush();
			}

			@fclose($handle);
		}
		else
		{
			@readfile($filepath);
			@flush();
		}

		// Ungraceful application exit -- so that any plugins won't screw up the download...
		$this->app->close();
	}

	/** @inheritdoc */
	public function getModel($name = 'Attachment', $prefix = 'Administrator', $config = ['ignore_request' => true])
	{
		return parent::getModel($name, $prefix, $config);
	}

	protected function onAfterDelete()
	{
		$this->applyReturnUrl();
	}

	protected function onAfterPublish()
	{
		$this->applyReturnUrl();
	}

	protected function onAfterUnpublish()
	{
		$this->applyReturnUrl();
	}

	public function checkToken($method = 'request', $redirect = true)
	{
		return parent::checkToken($method, $redirect);
	}
}