<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2023 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\Component\ATS\Administrator\Controller;

defined('_JEXEC') or die;

use Akeeba\Component\ATS\Administrator\Helper\Permissions;
use Akeeba\Component\ATS\Administrator\MailAuthentication\MailAuthenticationInterface;
use Akeeba\Component\ATS\Administrator\MailAuthentication\ServiceErrorException;
use Akeeba\Component\ATS\Administrator\Mixin\ControllerEventsTrait;
use Akeeba\Component\ATS\Administrator\Mixin\ControllerReusableModelsTrait;
use Akeeba\Component\ATS\Administrator\View\Authenticate\HtmlView;
use Exception;
use Joomla\CMS\Language\Text;
use Joomla\CMS\MVC\Controller\BaseController;
use RuntimeException;

/**
 * Controller for email services OAuth2 authentication
 *
 * @since  5.0.0
 */
class AuthenticateController extends BaseController
{
	use ControllerEventsTrait;
	use ControllerReusableModelsTrait;

	/**
	 * OAuth2 callback
	 *
	 * This is the third step of the OAuth2 authentication. It receives a code from the external service and exchanges
	 * it for an access and a refresh token which is the displayed to the user. THe user can then copy them over to the
	 * plugin page.
	 *
	 * @return  self
	 *
	 * @throws  Exception
	 * @since   5.0.0
	 */
	public function callback()
	{
		// Permissions check
		$identity = Permissions::getUser();

		if (
			!($identity->authorise('core.edit', 'com_plugins'))
			|| !($identity->authorise('core.manage', 'com_ats'))
		)
		{
			throw new RuntimeException(Text::_('JERROR_ALERTNOAUTHOR'), 403);
		}

		// I need the service in the query string
		$service = $this->input->getCmd('service');

		if (empty($service))
		{
			throw new RuntimeException(Text::_('JERROR_ALERTNOAUTHOR'), 403);
		}

		/** @var HtmlView $view */
		$view          = $this->getView();
		$view->service = $service;

		// Do we have errors?
		$error            = $this->input->getCmd('error', '');
		$errorDescription = $this->input->getRaw('error_description', '');

		if (!empty($error))
		{
			$view->error            = $error;
			$view->errorDescription = $errorDescription;

			$view->setLayout('error');

			return parent::display(false);
		}

		// Sanity check: I need client ID and api secret in the session
		$clientId = $this->app->getSession()->get('com_ats.auth.' . $service . 'client_id');
		$apiSecret = $this->app->getSession()->get('com_ats.auth.' . $service . 'api_secret');

		if (empty($clientId) || empty($apiSecret))
		{
			throw new RuntimeException(Text::_('JERROR_ALERTNOAUTHOR'), 403);
		}

		// Get the email service object
		$serviceObject = $this->getServiceObject($service);

		if (empty($serviceObject))
		{
			throw new RuntimeException(Text::_('JERROR_ALERTNOAUTHOR'), 403);
		}

		// We must have all expected incoming parameters (code and state)
		$code  = $this->input->getRaw('code', '');
		$state = $this->input->getRaw('state', '');

		if (empty($code) || empty($state))
		{
			throw new RuntimeException(Text::_('JERROR_ALERTNOAUTHOR'), 403);
		}

		// The state variable must match anti-CSRF token
		if ($state != $this->app->getFormToken())
		{
			throw new RuntimeException(Text::_('JERROR_ALERTNOAUTHOR'), 403);
		}

		// Exchange the code for a set of tokens
		/** @var HtmlView $view */
		$view = $this->getView();

		try
		{
			$tokens = $serviceObject->getTokens($code, $clientId, $apiSecret);

			$access_token  = $tokens['access'];
			$refresh_token = $tokens['refresh'];
		}
		catch (ServiceErrorException $e)
		{
			$info = json_decode($e->getMessage(), true);

			$view->error            = $info['error'] ?? 'generic';
			$view->errorDescription = $info['description'];

			$view->setLayout('error');

			if (isset($info['uri']))
			{
				$this->setRedirect($info['uri']);

				$this->redirect();

				return $this;
			}

			return parent::display(false);
		}

		$view->accessToken  = $access_token;
		$view->refreshToken = $refresh_token;

		$view->setLayout('default');

		return parent::display(false);
	}

	/**
	 * Initiate the OAuth2 flow.
	 *
	 * Redirects the user to the external mail service's OAuth2 consent page. The user will consent to the OAuth2 login.
	 * The service will then redirect them to our callback page to finish the login process.
	 *
	 * @return  self
	 *
	 * @throws  Exception
	 * @since   5.0.0
	 */
	public function consent()
	{
		// Anti-CSRF token check
		$this->checkToken('get');

		// Permissions check
		$identity = Permissions::getUser();

		if (
			!($identity->authorise('core.edit', 'com_plugins'))
			|| !($identity->authorise('core.manage', 'com_ats'))
		)
		{
			throw new RuntimeException(Text::_('JERROR_ALERTNOAUTHOR'), 403);
		}

		// Get and check required URL parameters
		$clientId  = $this->input->getRaw('client_id', '');
		$apiSecret = $this->input->getRaw('api_secret', '');
		$apiTenant = $this->input->getRaw('tenant', '');
		$service   = $this->input->getCmd('service', '');

		if (empty($clientId) || empty($apiSecret) || empty($service))
		{
			throw new RuntimeException(Text::_('JERROR_ALERTNOAUTHOR'), 403);
		}

		// Get the email service object
		$serviceObject = $this->getServiceObject($service);

		if (empty($serviceObject))
		{
			throw new RuntimeException(Text::_('JERROR_ALERTNOAUTHOR'), 403);
		}

		// Save Client ID and Secret Key in the session
		$this->app->getSession()->set('com_ats.auth.' . $service . 'client_id', $clientId);
		$this->app->getSession()->set('com_ats.auth.' . $service . 'api_secret', $apiSecret);
		$this->app->getSession()->set('com_ats.auth.' . $service . 'tenant', $apiTenant);

		$this->setRedirect($serviceObject->getRedirectionUrl($clientId));

		$this->redirect();

		return $this;
	}

	/** @inheritdoc */
	public function display($cachable = false, $urlparams = [])
	{
		// No default page for this view, sorry
		throw new RuntimeException(Text::_('JERROR_ALERTNOAUTHOR'), 403);
	}

	/**
	 * Get the mail service object or NULL if it does not exist
	 *
	 * @param   string  $service  The mail service name
	 *
	 * @return  MailAuthenticationInterface|null
	 *
	 * @since   5.0.0
	 */
	protected function getServiceObject(string $service): ?MailAuthenticationInterface
	{
		if (empty($service))
		{
			return null;
		}

		$fqn = '\\Akeeba\\Component\\ATS\\Administrator\\MailAuthentication\\' . ucfirst($service);

		if (!class_exists($fqn))
		{
			return null;
		}

		return new $fqn();
	}
}