<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2023 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\Component\ATS\Administrator\Controller;

defined('_JEXEC') or die;

use Akeeba\Component\ATS\Administrator\Helper\Permissions;
use Akeeba\Component\ATS\Administrator\Mixin\ControllerEventsTrait;
use Akeeba\Component\ATS\Administrator\Mixin\ControllerNewPostTrait;
use Akeeba\Component\ATS\Administrator\Mixin\ControllerReturnURLTrait;
use Akeeba\Component\ATS\Administrator\Mixin\ControllerReusableModelsTrait;
use Akeeba\Component\ATS\Administrator\Model\ManagernoteModel;
use Akeeba\Component\ATS\Administrator\Table\ManagernoteTable;
use Akeeba\Component\ATS\Administrator\Table\TicketTable;
use Akeeba\Component\ATS\Administrator\View\Post\HtmlView;
use Exception;
use Joomla\CMS\MVC\Controller\FormController;
use Joomla\CMS\MVC\Model\BaseDatabaseModel;

class ManagernoteController extends FormController
{
	use ControllerReusableModelsTrait;
	use ControllerEventsTrait;
	use ControllerReturnURLTrait;
	use ControllerNewPostTrait;

	/** @inheritdoc */
	protected $text_prefix = 'COM_ATS_MANAGERNOTE';

	public function edit($key = null, $urlVar = null)
	{
		// Joomla 4.1.1 and later will only allow cid as a POST variable. We need to use it with GET as well.
		$cid = (array) $this->input->get('cid', [], 'int');

		if (!empty($cid))
		{
			$this->input->post->set('cid', $cid);
		}

		return parent::edit($key, $urlVar);
	}

	protected function allowAdd($data = [])
	{
		$ticket_id = $data['ticket_id'] ?? null;

		if (empty($ticket_id))
		{
			return false;
		}

		/** @var TicketTable $ticket */
		$ticket = $this->getModel()->getTable('Ticket', 'Administrator');

		if (!$ticket->load($ticket_id))
		{
			return false;
		}

		$canDo = Permissions::getTicketPrivileges($ticket);

		return $canDo['admin'] || $canDo['notes.create'];
	}

	/** @inheritdoc */
	protected function allowEdit($data = [], $key = 'id')
	{
		$ticket_id = $data['ticket_id'] ?? null;
		$note_id   = $data['id'] ?? null;

		if (empty($ticket_id) && empty($note_id))
		{
			return false;
		}

		/** @var ManagernoteTable $note */
		$note = $this->getModel()->getTable('Managernote', 'Administrator');
		/** @var TicketTable $ticket */
		$ticket = $this->getModel()->getTable('Ticket', 'Administrator');

		if (!empty($note_id))
		{
			if (!$note->load($note_id))
			{
				return false;
			}
		}

		$ticket_id = $ticket_id ?: $note->ticket_id;

		if (empty($ticket_id))
		{
			return false;
		}

		if (!$ticket->load($ticket_id))
		{
			return false;
		}

		$canDo = Permissions::getTicketPrivileges($ticket);

		$canEditOwn = false;

		if ($note->created_by == Permissions::getUser()->id)
		{
			$canEditOwn = $canDo['notes.edit.own'];
		}

		return $canDo['admin'] || $canDo['notes.edit'] || $canEditOwn;
	}

	/**
	 * Runs after the save task (Save & Close button).
	 *
	 * Used to apply the return URL.
	 *
	 * @return  void
	 * @since   5.0.0
	 */
	protected function onAfterSave()
	{
		$this->applyReturnUrl();
	}

	/**
	 * Runs before the main display task.
	 *
	 * Used to communicate the return URL to the view.
	 *
	 * @return  void
	 * @throws  Exception
	 * @since   5.0.0
	 */
	protected function onBeforeMain(): void
	{
		/** @var HtmlView $view */
		$view            = $this->getView();
		$view->returnUrl = $this->getReturnUrl();
	}

	/**
	 * Called after we have finished saving a manager's note.
	 *
	 * Used to trigger notifications for manager notes, e.g. emails.
	 *
	 * @param   BaseDatabaseModel|ManagernoteModel  $model      The model this controller acts upon
	 * @param   array                               $validData  Validated data being saved.
	 *
	 * @return  void
	 * @throws  Exception
	 * @since   5.0.0
	 */
	protected function postSaveHook(BaseDatabaseModel $model, $validData = []): void
	{
		$noteId = $model->getState('managernote.id', $validData['id'] ?? null);

		if (!$noteId)
		{
			return;
		}

		/** @var ManagernoteTable $note */
		$note = $model->getItem();

		// Send post notifications (emails etc)
		$this->managerNoteNotifiable($note);
	}
}