<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2023 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\Component\ATS\Administrator\Controller;

defined('_JEXEC') or die;

use Akeeba\Component\ATS\Administrator\Mixin\ControllerEventsTrait;
use Akeeba\Component\ATS\Administrator\Mixin\ControllerReusableModelsTrait;
use Akeeba\Component\ATS\Administrator\Model\SchedulingModel;
use Exception;
use Joomla\CMS\Language\Text;
use Joomla\CMS\MVC\Controller\BaseController;
use RuntimeException;

class SchedulingController extends BaseController
{
	use ControllerEventsTrait;
	use ControllerReusableModelsTrait;

	/**
	 * Applies the randomly generated Secret Word in the session
	 *
	 * @return  void
	 *
	 * @throws  Exception
	 *
	 * @since   3.2.0
	 */
	public function applySecretWord()
	{
		$this->checkToken();

		/** @var SchedulingModel $model */
		$model = $this->getModel();
		$model->applySecretWord();

		$this->setRedirect('index.php?option=com_ats&view=scheduling');
	}

	/**
	 * The main display task.
	 *
	 * @return  self
	 *
	 * @throws  Exception
	 *
	 * @since   5.0.0
	 */
	public function display($cachable = false, $urlparams = [])
	{
		if ($this->input->getMethod() == 'POST')
		{
			$this->checkToken();

			$commandName = $this->input->post->getCmd('commandName', 'commandName');
			$commandName = empty($commandName) ? 'commandName' : $commandName;

			$this->app->getSession()->set('ats.cron.commandName', $commandName);
		}

		$this->conditionallyGenerateNewSecretWord();

		return parent::display($cachable, $urlparams);
	}

	/**
	 * Publish a plugin as public and set its Enable CRON option (if it exists) to 1
	 *
	 * @return  void
	 *
	 * @since   3.2.0
	 */
	public function fixPlugin()
	{
		$plugin = $this->input->get('plugin', '');

		/** @var SchedulingModel $model */
		$model = $this->getModel();
		$model->publishPlugin($plugin);

		$this->setRedirect('index.php?option=com_ats&view=scheduling');
	}

	/**
	 * Apply custom access control for all tasks.
	 *
	 * @param   string  $task  The task being executed
	 *
	 * @return  void
	 *
	 * @since   3.2.0
	 */
	protected function onBeforeExecute(&$task)
	{
		$me = $this->app->getIdentity();

		if (!$me->authorise('core.manage', 'com_ats'))
		{
			throw new RuntimeException(Text::_('JERROR_ALERTNOAUTHOR'), 403);
		}
	}

	/**
	 * Generate a new, random Secret Word if one is not set and we have not already stored one in the session.
	 *
	 * @return  void
	 *
	 * @since   3.2.0
	 */
	private function conditionallyGenerateNewSecretWord()
	{
		/** @var SchedulingModel $model */
		$model = $this->getModel();

		// Make sure we have a secret word
		if ($model->hasSecretWord())
		{
			return;
		}

		// If we don't have a secret word generate a new, random one but only once per session.
		$newSecretWord = $this->app->getSession()->get('ats.cron.newSecretWord', '');

		if (empty($newSecretWord))
		{
			$model->createNewSecretWord();
		}
	}
}