<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2023 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\Component\ATS\Administrator\Helper;

defined('_JEXEC') or die;

use Joomla\CMS\Factory;
use Joomla\CMS\User\User;
use Joomla\Database\DatabaseDriver;
use Joomla\Database\ParameterType;

abstract class AddonEmails
{
	/**
	 * Convert the stored values of add–on emails to an associative array (email => real_name, ...).
	 *
	 * There are two kinds of formats supported:
	 * 1. Joomla 3 repeatable field JSON format:
	 *      {"email":["client1@local.web","client2@local.web"],"name":["Client One","Client Two"]}
     * 2. Joomla 4 subform field JSON format:
	 *      {"__field10":{"email":"client1@local.web","name":"Client One"},"__field11":{"email":"client2@local.web","name":"Client 2"}}
     *
	 * In both cases the resulting array looks like this:
	 * [
	 *  'client1@local.web' => 'Client One',
	 *  'client2@local.web' => 'Client Two',
	 *  'client3@local.web' => 'Client Three',
	 * ]
	 *
	 * @param   string  $json
	 *
	 * @return  array
	 *
	 * @since   2.4.0
	 */
	public static function decodeAddonEmailJSON(string $json): array
	{
		$assoc = @json_decode($json, true);

		if (empty($assoc))
		{
			return [];
		}

		// Joomla 3 format decoding (for legacy data)
		if (isset($assoc['email']) && isset($assoc['name']))
		{
			$ret = [];
			$i   = 0;

			foreach ($assoc['email'] as $email)
			{
				$ret[$email] = $assoc['name'][$i];
				$i++;
			}

			return $ret;
		}

		$ret = [];

		foreach ($assoc as $fieldName => $info)
		{
			if (!isset($info['email']) || !isset($info['name']))
			{
				continue;
			}

			$ret[$info['email']] = $info['name'];
		}

		return $ret;
	}

	/**
	 * The inverse of decodeAddonEmailJSON.
	 *
	 * It takes the associative array of email => full name and encodes it back into the Joomla! subform JSON format.
	 *
	 * @param   array  $emailAssoc
	 *
	 * @return  string  JSON encoded string
	 *
	 * @since   2.4.0
	 */
	public static function encodeAddonEmailJSON(array $emailAssoc): string
	{
		$ret = [];
		$counter = 10;

		foreach ($emailAssoc as $email => $name)
		{
			$key = '__field' . $counter++;
			$ret[$key] = [
				'email' => $email,
				'name' => $name,
			];
		}

		return json_encode($ret);
	}

	/**
	 * Filters an associative list of email => full name. If the email exists as the main or add-on email of another
	 * user it will be removed from the list of emails. It's crucial to provide the correct $user_id value. Any records
	 * found to exist as add-on emails for a user _other_ than $user_id will be filtered out.
	 *
	 * @param   array  $emailAssoc  The list of emails to filter
	 * @param   int    $user_id     The user record these emails will be added to.
	 *
	 * @return  array  The filtered array
	 *
	 * @since   2.4.0
	 */
	public static function filterAddonEmails(array $emailAssoc, int $user_id): array
	{
		if (empty($emailAssoc))
		{
			return [];
		}

		$ret = [];

		foreach ($emailAssoc as $email => $name)
		{
			$id = self::findUserIdByAddonEmail($email);

			if ($id && ($id != $user_id))
			{
				continue;
			}

			$ret[$email] = $name;
		}

		return $ret;
	}

	/**
	 * Find the user ID given an add-on email.
	 *
	 * This method will first look for a user whose main email address matches $email. If none is found, it will look
	 * into the user profiles, doing a text match against the JSON data. The matched user records will then be loaded
	 * and processed properly. The user ID of the first record to legitimately include $email in the add-on email list
	 * will be returned.
	 *
	 * @param   string  $email  The email to search
	 *
	 * @return  int  The user ID or 0 if no match is found
	 *
	 * @since   2.4.0
	 */
	public static function findUserIdByAddonEmail(string $email): int
	{
		/** @var DatabaseDriver $db */
		$db          = Factory::getContainer()->get('DatabaseDriver');
		$email_lower = strtolower($email);
		$email_json  = json_encode($email_lower);

		// First look for an exact match in #__users
		$query = $db->getQuery(true)
			->select('id')
			->from($db->quoteName('#__users'))
			->where($db->quoteName('email') . ' = :email', 'OR')
			->where($db->quoteName('email') . ' = :emailLower', 'OR')
			->bind(':email', $email)
			->bind(':emailLower', $email_lower);
		$id    = $db->setQuery($query)->loadResult();

		if (!empty($id))
		{
			return $id;
		}

		// Then look for add-on emails in the user profiles
		$likeEmailJson = '%' . $email_json . '%';
		$key           = 'ats.notification_emails';
		$query         = $db->getQuery(true)
			->select('user_id')
			->from($db->quoteName('#__user_profiles'))
			->where($db->quoteName('profile_key') . ' = :key')
			->where($db->quoteName('profile_value') . ' LIKE :like')
			->bind(':key', $key)
			->bind(':like', $likeEmailJson);

		$userIDs = $db->setQuery($query)->loadColumn(0);

		// No match? Return 0.
		if (empty($userIDs))
		{
			return 0;
		}

		// Finally make sure we really did find the exact add-on email in a user profile record, or return 0
		foreach ($userIDs as $uid)
		{
			$emails = self::getAddonEmails($uid);

			if (array_key_exists($email, $emails))
			{
				return $uid;
			}
		}

		return 0;
	}

	/**
	 * Get a list of add-on emails for the given user ID. The emails are returned as an associative array of email =>
	 * full name.
	 *
	 * @param   int|null  $user_id  The numeric ID of the user
	 *
	 * @return  array Something like ['user1@example.com' => 'User One', 'user2@example.com' => 'User Two', ...]
	 *
	 * @since   2.4.0
	 */
	public static function getAddonEmails(?int $user_id): array
	{
		if (empty($user_id))
		{
			return [];
		}

		/** @var DatabaseDriver $db */
		$db = Factory::getContainer()->get('DatabaseDriver');

		$key   = 'ats.notification_emails';
		$query = $db->getQuery(true)
			->select('profile_value')
			->from($db->quoteName('#__user_profiles'))
			->where($db->quoteName('profile_key') . ' = :key')
			->where($db->quoteName('user_id') . ' = :user_id')
			->bind(':key', $key)
			->bind(':user_id', $user_id, ParameterType::INTEGER);

		$json = $db->setQuery($query)->loadResult();

		if (empty($json))
		{
			return [];
		}

		return self::decodeAddonEmailJSON($json);
	}

	/**
	 * Get a list of add-on emails for the given user object.
	 *
	 * The emails are returned as an associative array of email => full name.
	 *
	 * @param   User  $user  The user object
	 *
	 * @return  array
	 *
	 * @since   2.4.0
	 */
	public static function getAddonEmailsByUser(User $user): array
	{
		return self::getAddonEmails($user->id);
	}
}