<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2023 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\Component\ATS\Administrator\Helper;

defined('_JEXEC') or die;

use Akeeba\Component\ATS\Administrator\Table\AttachmentTable;
use Akeeba\Component\ATS\Administrator\Table\PostTable;
use Exception;
use Joomla\CMS\Component\ComponentHelper;
use Joomla\CMS\Factory;
use Joomla\CMS\Filesystem\File;
use Joomla\CMS\Filesystem\Folder;
use Joomla\CMS\Filesystem\Path as JPath;
use Joomla\CMS\Filter\InputFilter;
use Joomla\CMS\Helper\MediaHelper;
use Joomla\CMS\Language\Text;
use RuntimeException;

/**
 * A helper class to manage attachments
 *
 * @since  5.0.0
 */
class Attachment
{
	/**
	 * The absolute filesystem path where attachments are stored.
	 *
	 * @var   null|string
	 * @since 5.0.0
	 */
	private static $attachmentsDirectory = null;

	/**
	 * Returns the absolute filesystem path where attachments are stored.
	 *
	 * @return  string
	 *
	 * @since   5.0.0
	 */
	public static function getDirectory(): string
	{
		if (!empty(self::$attachmentsDirectory))
		{
			return self::$attachmentsDirectory;
		}

		$cParams   = ComponentHelper::getParams('com_ats');
		$directory = $cParams->get('attachments_folder', 'media/com_ats/attachments');
		$directory = trim($directory);
		$directory = rtrim($directory, '/\\');

		// Empty setting will result in the default being used
		if (empty($directory))
		{
			self::$attachmentsDirectory = JPATH_SITE . '/media/com_ats/attachments';

			return self::$attachmentsDirectory;
		}

		// Is this a subdirectory of the site's root (as it should be)?
		try
		{
			$possiblePath = JPATH_ROOT . $directory;
			$cleanDir     = JPath::check($possiblePath);

			if (is_dir($cleanDir))
			{
				self::$attachmentsDirectory = $cleanDir;

				return $cleanDir;
			}
		}
		catch (Exception $e)
		{
			// Nope. Let's fall through.
		}

		// Everything failed, return the default.
		self::$attachmentsDirectory = JPATH_SITE . '/media/com_ats/attachments';

		return self::$attachmentsDirectory;
	}

	/**
	 * Process the uploaded files and attach them to a ticket.
	 *
	 * @param   array      $files  Uploaded files, as pre–processed by Joomla's FileInput
	 * @param   PostTable  $post   The post to attach the uploads to.
	 *
	 * @return  array{attachment_id:string,errors:array}
	 */
	public static function manageUploads(array $files, PostTable $post): array
	{
		if (empty($files))
		{
			return [
				'attachment_id' => '',
				'errors'        => [],
			];
		}

		// Filter out invalid records
		$files = array_filter($files, function ($file) {
			return is_array($file) && !empty($file) && array_key_exists('name', $file);
		});

		// Keep up to 10 attachments (hard limit)
		$files = array_slice($files, 0, 10);

		// Convert filenames to something Joomla's media manager will accept
		$files = array_map(function (array $file) {
			$file['name'] = File::makeSafe($file['name']);

			return $file;
		}, $files);

		$errors = [];
		$return = [];

		foreach ($files as $file)
		{
			try
			{
				$filedef = self::uploadFile($file, true);
			}
			catch (Exception $e)
			{
				$errors[] = Text::sprintf('COM_ATS_ATTACHMENTS_ERR_UPLOAD', $file['name'], $e->getMessage());

				continue;
			}

			$attachmentTable = new AttachmentTable(Factory::getContainer()->get('DatabaseDriver'));
			$attachmentTable->reset();

			$filedef['post_id']    = $post->id;
			$filedef['created_by'] = Permissions::getUser()->id;
			$filedef['enabled']    = 1;

			if (!$attachmentTable->save($filedef))
			{
				$errors[] = Text::sprintf('COM_ATS_ATTACHMENTS_ERR_UPLOAD', $file['name'], $attachmentTable->getError());

				continue;
			}

			$return[] = $attachmentTable->getId();
		}

		return [
			'attachment_id' => implode(',', $return),
			'errors'        => $errors,
		];
	}

	/**
	 * Secure the attachments directory.
	 *
	 * @return  void
	 *
	 * @since   5.0.0
	 */
	public static function secureAttachmentsDirectory(): void
	{
		$dir = self::getDirectory();

		// Create a .htaccess file to prevent all web access (Apache 1.3+, Lightspeed, Lighttpd, ...)
		if (!is_file($dir . '/.htaccess'))
		{
			$htaccess = <<< APACHE
<IfModule !mod_authz_core.c>
Order deny,allow
Deny from all
</IfModule>
<IfModule mod_authz_core.c>
  <RequireAll>
    Require all denied
  </RequireAll>
</IfModule>
APACHE;

			@file_put_contents($dir . '/.htaccess', $htaccess);
		}

		// Create a web.config to prevent all web access (IIS 7+)
		if (!is_file($dir . '/web.config'))
		{
			$webConfig = <<< XML
<?xml version="1.0"?>
<!--
    This only works on IIS 7 or later. See https://www.iis.net/configreference/system.webserver/security/requestfiltering/fileextensions
-->
<configuration>
    <system.webServer>
        <security>
            <requestFiltering>
                <fileExtensions allowUnlisted="false" >
                    <clear />
                    <add fileExtension=".html" allowed="true"/>
                </fileExtensions>
            </requestFiltering>
        </security>
    </system.webServer>
</configuration>
XML;
			@file_put_contents($dir . '/web.config', $webConfig);
		}

		// Create a blank index.html or index.htm to prevent directory listings (all servers)
		$blankHtml = <<< HTML
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
<html>
  <head>
    <title>Access Denied</title>
  </head>
  <body>
	  <h1>Access Denied</h1>
  </body>
</html>
HTML;

		if (!is_file($dir . '/index.html'))
		{
			@file_put_contents($dir . '/index.html', $blankHtml);
		}

		if (!is_file($dir . '/index.htm'))
		{
			@file_put_contents($dir . '/index.htm', $blankHtml);
		}

		// Create a default index.php to prevent directory listings with an error (all servers)
		if (!is_file($dir . '/index.php'))
		{
			$deadPHP = '<' . '?' . 'php header(\'HTTP/1.1 403 Forbidden\');';
			@file_put_contents($dir . '/index.php', $deadPHP);
		}
	}

	/**
	 * Finalises the upload of an attachment into the attachments directory
	 *
	 * Moves an uploaded file to the configured attachments directory under a random name and returns a full file
	 * definition array, or false if the upload failed for any reason.
	 *
	 * @param   array  $file         The file descriptor returned by PHP
	 * @param   bool   $checkUpload  Should I check if upload is allowed for this file?
	 *
	 * @return  array
	 *
	 * @throws  RuntimeException|Exception
	 */
	public static function uploadFile(array $file, bool $checkUpload = true): array
	{
		if (!isset($file['name']))
		{
			throw new RuntimeException(Text::_('COM_ATS_ATTACHMENTS_ERR_NOFILE'));
		}

		$cParams       = ComponentHelper::getParams('com_ats');
		$unsafeUploads = $cParams->get('unsafe_uploads', 0) == 1;
		$app           = Factory::getApplication();

		if (!$unsafeUploads)
		{
			self::checkUpload($file);
		}

		// Get a (very!) randomised name
		$serverkey = $app->get('secret', '');
		$sig       = $file['name'] . microtime() . $serverkey;

		if (function_exists('sha256'))
		{
			$mangledname = sha256($sig);
		}
		elseif (function_exists('sha1'))
		{
			$mangledname = sha1($sig);
		}
		else
		{
			$mangledname = md5($sig);
		}

		// ...and its full path
		$relPath  = substr($mangledname, 0, 2) . '/' . substr($mangledname, 2, 2) . '/';
		$filepath = JPath::clean(self::getDirectory() . '/' . $relPath . $mangledname);

		$dirName = dirname($filepath);

		if (!@is_dir($dirName) && !@mkdir($dirName, 0755, true))
		{
			Folder::create($filepath);
		}

		// If we have a name clash, abort the upload
		if (File::exists($filepath))
		{
			throw new Exception(Text::_('COM_ATS_ATTACHMENTS_ERR_NAMECLASH'));
		}

		// Do the upload
		if ($checkUpload)
		{
			if (!File::upload($file['tmp_name'], $filepath))
			{
				throw new Exception(Text::_('COM_ATS_ATTACHMENTS_ERR_CANTJFILEUPLOAD'));
			}
		}
		else
		{
			$copied = @copy($file['tmp_name'], $filepath);

			if (!$copied)
			{
				$copied = File::copy($file['tmp_name'], $filepath);
			}

			if (!$copied)
			{
				throw new Exception(Text::_('COM_ATS_ATTACHMENTS_ERR_CANTJFILEUPLOAD'));
			}
		}

		// Get the MIME type
		$mime = 'application/octet-stream';

		if (function_exists('mime_content_type'))
		{
			$mime = mime_content_type($filepath);
		}
		elseif (function_exists('finfo_open'))
		{
			$finfo = finfo_open(FILEINFO_MIME_TYPE);
			$mime  = finfo_file($finfo, $filepath);
		}

		// Return the file info
		return [
			'original_filename' => $file['name'],
			'mangled_filename'  => $mangledname,
			'mime_type'         => $mime,
		];
	}

	/**
	 * Performs security checks on the upload file
	 *
	 * @param   array  $file
	 *
	 * @return  void
	 *
	 * @throws  Exception
	 * @since   5.0.0
	 */
	private static function checkUpload(array $file)
	{
		$app = Factory::getApplication();

		// Can we upload this file type?
		$paths = [JPATH_ROOT, JPATH_ADMINISTRATOR];

		$jlang = Factory::getApplication()->getLanguage();
		$jlang->load('com_media', $paths[0], 'en-GB', true);
		$jlang->load('com_media', $paths[0], null, true);
		$jlang->load('com_media', $paths[1], 'en-GB', true);
		$jlang->load('com_media', $paths[1], null, true);

		$mediaHelper = new MediaHelper();

		$allowedExecutables = ComponentHelper::getParams('com_ats')->get(
			'allowed_executables',
			implode(',', array_merge(MediaHelper::EXECUTABLES, InputFilter::FORBIDDEN_FILE_EXTENSIONS))
		);
		$allowedExecutables = array_map(function ($x) {
			return ltrim(trim($x), '.');
		}, explode(',', $allowedExecutables));

		if ($mediaHelper->canUpload($file, 'com_ats', $allowedExecutables))
		{
			return;
		}

		$errors = $app->getMessageQueue();
		$error = '';

		if (count($errors))
		{
			$error = array_pop($errors);
			$error  = $error['message'];
		}

		if (!empty($error))
		{
			throw new RuntimeException(Text::sprintf('COM_ATS_ATTACHMENTS_ERR_MEDIAHELPERERROR', $error));
		}

		throw new RuntimeException(Text::_('COM_ATS_POSTS_ERR_ATTACHMENTERRORGENERIC'));

	}
}