<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2023 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\Component\ATS\Administrator\Helper;

use Akeeba\Component\ATS\Administrator\Table\AutoreplyTable;
use Akeeba\Component\ATS\Administrator\Table\PostTable;
use Akeeba\Component\ATS\Administrator\Table\TicketTable;
use Exception;
use Joomla\CMS\Factory;
use Joomla\CMS\Language\Text;
use Joomla\CMS\Plugin\PluginHelper;
use Joomla\CMS\User\UserHelper;
use Joomla\Database\DatabaseDriver;
use Joomla\Database\ParameterType;
use Joomla\Registry\Registry;
use Joomla\Utilities\ArrayHelper;
use RuntimeException;

defined('_JEXEC') or die;

class AutoReply
{
	protected static $bots = null;

	/**
	 * Run the auto-reply rules and return the number of tickets we replied to.
	 *
	 * If $tickets is set to a non-empty array or Traversable object then only these tickets will be taken into account,
	 * without ticket age limits, without querying the database for tickets in need of an automatic reply.
	 *
	 * @param   int[]|null  $tickets  (array of integers) Only auto-reply to these ticket IDs
	 * @param   Timer|null  $timer    A timer object to pace ourselves
	 *
	 * @return  int  The number of tickets we replied to
	 */
	public static function runAutoreplies(?array $tickets = null, ?Timer $timer = null)
	{
		// Sanity check: are there any bot users?
		if (count(self::getBots()) == 0)
		{
			throw new RuntimeException(Text::_('COM_ATS_AUTOREPLIES_ERR_NOBOTS'));
		}

		// Let's get all the enabled autoreply rules
		/** @var DatabaseDriver $db */
		$db        = Factory::getContainer()->get('DatabaseDriver');
		$query     = $db->getQuery(true)
			->select('*')
			->from($db->quoteName('#__ats_autoreplies'))
			->where($db->quoteName('enabled') . ' = 1');
		$ruleTable = new AutoreplyTable($db);

		/** @var AutoreplyTable[] $rules */
		$rules = array_map(function ($data) use ($ruleTable) {
			$ret = clone $ruleTable;
			$ret->reset();
			$ret->bind($data);

			return $ret;
		}, $db->setQuery($query)->loadAssocList() ?: []);

		// Sanity check: are there any rules?
		if (count($rules) == 0)
		{
			return 0;
		}

		if (!is_object($timer) || !($timer instanceof Timer))
		{
			$timer = null;
		}

		$processedTickets = [];

		// And now let's start looking for suitable tickets
		foreach ($rules as $rule)
		{
			if (is_object($timer) && ($timer instanceof Timer))
			{
				if ($timer->getTimeLeft() <= 0.01)
				{
					break;
				}
			}

			// If I got an error let's bubble it out
			$theseTickets     = self::runRule($rule, $tickets, $timer, $processedTickets);
			$processedTickets = array_unique(array_merge($processedTickets, $theseTickets));
		}

		// If everything was ok, let's return the number of processed tickets
		return count($processedTickets);
	}

	/**
	 * Get the user IDs of the reply bot users
	 *
	 * @return  array
	 *
	 * @throws  Exception
	 * @since   5.0.0
	 */
	protected static function getBots(): array
	{
		if (!is_null(self::$bots))
		{
			return self::$bots;
		}

		self::$bots = [];

		// Let's check if I have any username
		$plugin = PluginHelper::getPlugin('ats', 'autoreply');

		if (!$plugin)
		{
			return self::$bots;
		}

		$pluginParams = new Registry($plugin->params);
		$rawData      = trim($pluginParams->get('botusers', ''));
		$rawData      = empty($rawData) ? '' : str_replace("\\n", "\n", $rawData);
		$rawData      = empty($rawData) ? '' : str_replace("\r\n", "\n", $rawData);
		$rawData      = empty($rawData) ? '' : str_replace("\r", "\n", $rawData);

		if (empty($rawData))
		{
			return self::$bots;
		}

		// Just in case something funky happened...
		$bots = array_map(function ($botUser) {
			if (empty($botUser))
			{
				return null;
			}

			$userId = UserHelper::getUserId($botUser);

			if (empty($userId))
			{
				return null;
			}

			$user = Permissions::getUser($userId);

			if (empty($user))
			{
				return null;
			}

			if (empty($user->id))
			{
				return null;
			}

			return $user->id;
		}, explode("\n", $rawData));

		self::$bots = array_filter($bots, function ($uID) {
			return !is_null($uID) && ($uID > 0);
		});

		return self::$bots;
	}

	/**
	 * Get applicable tickets for automatic replies
	 *
	 * @param   int[]  $categoryIDs       Categories to find tickets to reply to
	 * @param   int[]  $processedTickets  Tickets already replied to in this run
	 *
	 * @return  int[]  The ticket IDs automatic replies should be considered for
	 *
	 * @throws  Exception
	 * @since   5.0.0
	 */
	protected static function getTickets(array $categoryIDs = [], array $processedTickets = []): array
	{
		static $tickets = [];

		/** @var DatabaseDriver $db */
		$db = Factory::getContainer()->get('DatabaseDriver');

		// Remove any empty value
		$categoryIDs = array_diff($categoryIDs, ['', '0']);
		$hash        = md5(implode('', $categoryIDs));

		if (!isset($tickets[$hash]))
		{
			$bots = self::getBots();

			// Get the ticket id where a bot already replied in previous iterations
			$subQuery = $db->getQuery(true)
				->select($db->quoteName('ticket_id'))
				->from($db->quoteName('#__ats_posts'))
				->whereIn($db->quoteName('created_by'), $bots, ParameterType::INTEGER);

			// Get only open tickets
			$query = $db->getQuery(true)
				->select($db->quoteName('id'))
				->from($db->quoteName('#__ats_tickets'))
				->where($db->quoteName('status') . ' = ' . $db->quote('O'))
				->where($db->quoteName('id') . ' NOT IN(' . $subQuery . ')');

			if ($categoryIDs)
			{
				$query->whereIn($db->quoteName('catid'), $categoryIDs, ParameterType::INTEGER);
			}

			$tickets[$hash] = $db->setQuery($query)->loadColumn();
		}

		// Subtracts the already processed tickets in order to avoid double posting
		$tickets[$hash] = array_diff($tickets[$hash], $processedTickets);

		return $tickets[$hash];
	}

	/**
	 * Convert a list of keywords to an array.
	 *
	 *
	 * @param   string  $keywords  The list of keywords, separated by newlines
	 *
	 * @return  string[]
	 *
	 * @since   5.0.0
	 */
	protected static function keywordsListToArray($keywords)
	{
		$keywords = trim($keywords);
		$keywords = str_replace(["\\n", "\r", "\r\n", "\n\n"], ["\n", "\n", "\n", "\n"], $keywords);

		if ($keywords)
		{
			$ret = explode("\n", $keywords);

			return array_map('trim', $ret);
		}

		return [];
	}

	/**
	 * Runs a specific auto-reply rule
	 *
	 * If $tickets is set to a non-empty array or Traversable object then only these tickets will be taken into account,
	 * without ticket age limits, without querying the database for tickets in need of an automatic reply.
	 *
	 * @param   AutoreplyTable  $rule
	 * @param   int[]|null      $tickets  (array of integers) Only auto-reply to these ticket IDs
	 * @param   Timer|null      $timer    A timer object to pace ourselves
	 *
	 * @return  int[]  The IDs of the tickets processed
	 */
	protected static function runRule(AutoreplyTable $rule, ?array $tickets = null, ?Timer $timer = null, ?array $processedTickets): array
	{
		$categories     = [];
		$keywords_title = self::keywordsListToArray($rule->keywords_title);
		$keywords_text  = self::keywordsListToArray($rule->keywords_text);
		$ruleId         = $rule->getId();
		$params         = $rule->params;

		if (isset($params['categories']))
		{
			$categories = $params['categories'];
		}

		// $tickets could be null or an empty array. In either cases, we need to fetch the tickets from the cats
		$needsTickets = is_null($tickets);
		$needsTickets = $needsTickets || (is_array($tickets) && empty($tickets));

		if ($needsTickets)
		{
			$tickets = self::getTickets($categories, $processedTickets);
		}

		// Performs several checks vs the rule, narrowing the number of affected tickets
		$tickets = self::checkMinAfter($rule->min_after, $tickets);
		$tickets = self::checkNumPosts($rule->num_posts, $tickets);
		$tickets = self::checkAttachments($rule->attachment, $tickets);
		$tickets = self::checkManagerReply($rule->run_after_manager, $tickets);
		$tickets = self::checkTitleKeywords($keywords_title, $tickets);
		$tickets = self::checkTextKeywords($keywords_text, $tickets);

		if (empty($tickets))
		{
			return [];
		}

		$jNow = clone Factory::getDate();

		/** @var DatabaseDriver $db */
		$db     = Factory::getContainer()->get('DatabaseDriver');
		$post   = new PostTable($db);
		$ticket = new TicketTable($db);
		$bots   = self::getBots();

		foreach ($tickets as $ticketId)
		{
			if (is_object($timer) && ($timer instanceof Timer))
			{
				if ($timer->getTimeLeft() <= 0.01)
				{
					break;
				}
			}

			$ticket->reset();

			if (!$ticket->load($ticketId))
			{
				continue;
			}

			// Do I have any posts on this ticket from this auto-reply rule?
			$params      = $ticket->params;
			$autoReplies = $params->get('__ats__autoreplies', []);

			if (!is_array($autoReplies))
			{
				$autoReplies = [];
			}

			if (in_array($ruleId, $autoReplies))
			{
				continue;
			}

			$autoReplies[] = $ruleId;
			$params->set('__ats__autoreplies', $autoReplies);

			$current_bot = $bots[array_rand($bots)];

			/**
			 * I'll leave empty the creation date so that the table populates it automatically. Otherwise it will
			 * populate the modified_ column set, resulting in the post appearing edited.
			 */
			$data = [
				'ats_post_id'   => 0,
				'ats_ticket_id' => $ticketId,
				'content_html'  => $rule->reply,
				'created'       => null,
				'created_by'    => $current_bot,
				'enabled'       => 1,
			];

			$post->save($data);

			$ticket->params = $params;

			TicketTable::$noPermissionsCheck = true;

			$ticket->save([
				'status'      => 'P',
				'modified'    => $jNow->toSql(),
				'modified_by' => $current_bot,
			]);

			TicketTable::$noPermissionsCheck = false;

			$processedTickets[] = $ticketId;
		}

		return array_unique($processedTickets);
	}

	private static function checkAttachments(int $attach, array $tickets): array
	{
		/**
		 * If the rule applies to posts with or without attachments (ie every ticket), let's return the passed ticket
		 * array.
		 */
		if ($attach == 2 || empty($tickets))
		{
			return $tickets;
		}

		/** @var DatabaseDriver $db */
		$db = Factory::getContainer()->get('DatabaseDriver');

		$query = $db->getQuery(true)
			->select([
				$db->quoteName('p.ticket_id'),
				'SUM(' . $db->quoteName('a.id') . ') AS ' . $db->quoteName('attachments'),
			])
			->from($db->quoteName('#__ats_attachments', 'a'))
			->innerJoin($db->quoteName('#__ats_posts', 'p'),
				$db->quoteName('p.id') . ' = ' . $db->quoteName('a.post_id')
			)
			->whereIn($db->quoteName('p.ticket_id'), $tickets, ParameterType::INTEGER)
			->group($db->quoteName('p.ticket_id'));

		// I want tickets where AT LEAST ONE post has an attachment
		if ($attach)
		{
			$query->having($db->quoteName('attachments') . ' >= 0');
		}
		// I want tickets where NO POSTS have an attachments
		else
		{
			$query->having($db->quoteName('attachments') . ' = 0');
		}

		return $db->setQuery($query)->loadColumn() ?: [];
	}

	private static function checkManagerReply(int $manager, array $tickets): array
	{
		// Rule applies regardless the manager reply, let's return the passed ticket array
		if ($manager == 2 || empty($tickets))
		{
			return $tickets;
		}

		/** @var DatabaseDriver $db */
		$db       = Factory::getContainer()->get('DatabaseDriver');
		$managers = [];

		// Let's get all the affected categories
		$query = $db->getQuery(true)
			->select($db->quoteName('catid'))
			->from($db->quoteName('#__ats_tickets'))
			->whereIn($db->quoteName('id'), $tickets, ParameterType::INTEGER)
			->group($db->quoteName('catid'));

		$cats = $db->setQuery($query)->loadColumn();

		/**
		 * Now let's get all the managers.
		 *
		 * The only problem that could arise, is if a manager posts as customer inside another category he's not a
		 * manager. I really think this is a remote chance
		 */
		foreach ($cats as $cat)
		{
			$managers = array_merge($managers, Permissions::getManagers($cat));
		}

		$managers = ArrayHelper::getColumn($managers, 'id');

		$query = $db->getQuery(true)
			->select($db->quoteName('ticket_id'))
			->from($db->quoteName('#__ats_posts'))
			->whereIn($db->quoteName('ticket_id'), $tickets, ParameterType::INTEGER)
			->group($db->quoteName('ticket_id'));

		if (!$manager)
		{
			// I want only ticket where the manager didn't reply
			$query->whereNotIn($db->quoteName('created_by'), $managers, ParameterType::INTEGER);
		}
		else
		{
			// Only where where the manager replied
			$query->whereIn($db->quoteName('created_by'), $managers, ParameterType::INTEGER);
		}

		return $db->setQuery($query)->loadColumn() ?: [];

	}

	private static function checkMinAfter(int $minutes, array $tickets): array
	{
		if (empty($minutes) || ((int) $minutes <= 0) || empty($tickets))
		{
			return $tickets;
		}

		/** @var DatabaseDriver $db */
		$db   = Factory::getContainer()->get('DatabaseDriver');
		$time = (clone Factory::getDate('-' . $minutes . ' minutes'))->toSql();

		$query = $db->getQuery(true)
			->select($db->quoteName('id'))
			->from($db->quoteName('#__ats_tickets'))
			->whereIn($db->quoteName('id'), $tickets, ParameterType::INTEGER)
			->where($db->quoteName('modified') . ' <= :time')
			->bind(':time', $time);

		return $db->setQuery($query)->loadColumn() ?: [];
	}

	private static function checkNumPosts(int $numPost, array $tickets): array
	{
		// If I have no constraints on the number of posts, let's return the passed ticket array
		if (empty($numPost) || $numPost <= 1 || empty($tickets))
		{
			return $tickets;
		}

		/** @var DatabaseDriver $db */
		$db = Factory::getContainer()->get('DatabaseDriver');

		$query = $db->getQuery(true)
			->select([
				$db->quoteName('ticket_id'),
				'COUNT(*) AS ' . $db->quoteName('total'),
			])
			->from($db->quoteName('#__ats_posts'))
			->whereIn($db->quoteName('ticket_id'), $tickets, ParameterType::INTEGER)
			->group($db->quoteName('ticket_id'))
			->having($db->qn('total') . ' >= :numPost')
			->bind(':numPost', $numPost, ParameterType::INTEGER);

		return $db->setQuery($query)->loadColumn() ?: [];
	}

	private static function checkTextKeywords(array $keywords, array $tickets): array
	{
		if (empty($keywords) || empty($tickets))
		{
			return $tickets;
		}

		/** @var DatabaseDriver $db */
		$db    = Factory::getContainer()->get('DatabaseDriver');
		$parts = [];

		$query = $db->getQuery(true)
			->select($db->quoteName('id'))
			->from($db->quoteName('#__ats_tickets'))
			->whereIn($db->quoteName('id'), $tickets, ParameterType::INTEGER);

		$i = 0;
		foreach ($keywords as $keyword)
		{
			$varName  = 'keyword' . $i++;
			$$varName = '%' . $keyword . '%';

			$parts[] = $db->quoteName('content_html') . ' LIKE :' . $varName;
			$query->bind(':' . $varName, $$varName);
		}

		$query->extendWhere('AND', $parts, 'OR');

		return $db->setQuery($query)->loadColumn() ?: [];
	}

	private static function checkTitleKeywords(array $keywords, array $tickets): array
	{
		if (empty($keywords) || empty($tickets))
		{
			return $tickets;
		}

		/** @var DatabaseDriver $db */
		$db    = Factory::getContainer()->get('DatabaseDriver');
		$parts = [];

		$query = $db->getQuery(true)
			->select($db->quoteName('id'))
			->from($db->quoteName('#__ats_tickets'))
			->whereIn($db->quoteName('id'), $tickets, ParameterType::INTEGER);

		$i = 0;
		foreach ($keywords as $keyword)
		{
			$varName  = 'keyword' . $i++;
			$$varName = '%' . $keyword . '%';

			$parts[] = $db->quoteName('title') . ' LIKE :' . $varName;
			$query->bind(':' . $varName, $$varName);
		}

		$query->extendWhere('AND', $parts, 'OR');

		return $db->setQuery($query)->loadColumn() ?: [];
	}
}