<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2023 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\Component\ATS\Administrator\Helper;

use Akeeba\Component\ATS\Administrator\Exceptions\CronCommandMissing;
use Akeeba\Component\ATS\Administrator\Exceptions\CronCommandNotFound;
use Akeeba\Component\ATS\Administrator\Exceptions\SecretMismatch;
use Akeeba\Component\ATS\Administrator\Exceptions\SecretNotConfigured;
use Joomla\CMS\Application\CMSApplication;
use Joomla\CMS\Application\ConsoleApplication;
use Joomla\CMS\Component\ComponentHelper;
use Joomla\CMS\Crypt\Crypt;
use Joomla\CMS\Factory;
use Joomla\CMS\Log\Log;
use Joomla\CMS\Plugin\PluginHelper;

defined('_JEXEC') or die;

/**
 * CRON job execution helper
 *
 * @since  3.2.0
 */
abstract class Cron
{
	/**
	 * ATS CRON log filename
	 *
	 * @since 3.2.0
	 */
	const logFilename = 'ats_cron.php';

	/**
	 * Have I already installed a Joomla logger method for the 'ats.cron' context?
	 *
	 * @var   bool
	 * @since 3.2.0
	 */
	private static $loggerInstalled = false;

	/**
	 * Record an entry in the ATS CRON log file.
	 *
	 * The log file is in the site's log directory. Its name is in the logFilename constant in this class.
	 *
	 * @param   string  $entry
	 * @param   int     $level
	 *
	 * @return  void
	 * @since   3.2.0
	 */
	public static function log($entry, $level = Log::DEBUG)
	{
		// Install a logger for ATS CRON jobs if necessary
		if (!self::$loggerInstalled)
		{
			Logger::registerFileLogger('ats.cron');

			self::$loggerInstalled = true;
		}

		Log::add($entry, $level, 'ats.cron');
	}

	/**
	 * Tries to execute the given command. Returns true if the command executed successfully.
	 *
	 * @param   string  $command    The command to execute. The filtered command name is written back to this var.
	 * @param   int     $timeLimit  Time limit in seconds. 0 means "no limit", is implemented as a 86400s (24h) limit.
	 *
	 * @return  bool  True on successful execution
	 *
	 * @since   3.2.0
	 */
	public static function run(&$command, $timeLimit = 0)
	{
		// Perform more rigorous filtering of the command name
		$command = (string) preg_replace('/[^A-Z0-9_]/i', '', $command);
		$command = strtolower($command);

		if (empty($command))
		{
			throw new CronCommandMissing();
		}

		// Default processing time limit: a ridiculous 24 hours
		$timeLimit = 86400;

		// Process the time limit
		if ($timeLimit > 0)
		{
			// If a time limit is set clamp it between 1 and 600 seconds
			$timeLimit = max($timeLimit, 1);
			$timeLimit = min($timeLimit, 600);
		}

		self::log(sprintf('Preparing to execute command “%s”', $command));

		// Make sure the language files are loaded
		/** @var CMSApplication|ConsoleApplication $app */
		$app      = Factory::getApplication();
		$language = $app->getLanguage();
		$isAdmin  = $app->isClient('administrator');
		$thisPath = $isAdmin ? JPATH_ADMINISTRATOR : JPATH_ROOT;
		$altPath  = !$isAdmin ? JPATH_ADMINISTRATOR : JPATH_ROOT;
		$language->load('com_ats', $altPath, null, true);
		$language->load('com_ats', $thisPath, null, true);

		// Run the command
		PluginHelper::importPlugin('ats');
		$result = $app->triggerEvent('onAtsCronTask', [
			$command,
			[
				'time_limit' => $timeLimit,
			],
		]);

		$commandFound = array_reduce($result, function ($carry, $item) {
			if ($carry)
			{
				return true;
			}

			return !is_null($item);
		}, false);

		if (!$commandFound)
		{
			throw new CronCommandNotFound();
		}

		$finalResult = array_reduce($result, function ($carry, $item) {
			if (is_null($item))
			{
				return $carry;
			}

			return $item;
		}, false);

		return (bool) $finalResult;
	}

	/**
	 * Checks if the provided secret key matches the configuration.
	 *
	 * The comparison is performed in a timing safe manner. An exception is thrown if the key does not match or if no
	 * secret key has been configured yet
	 *
	 * @param   string  $key  The secret key to check.
	 *
	 * @return  void
	 *
	 * @throws  SecretNotConfigured
	 * @throws  SecretMismatch
	 *
	 * @since   3.2.0
	 */
	public static function checkSecret($key)
	{
		$cParams          = ComponentHelper::getParams('com_ats');
		$configuredSecret = trim($cParams->get('secret', ''));
		$key              = trim($key);

		if (empty($configuredSecret))
		{
			throw new SecretNotConfigured();
		}

		if (!Crypt::timingSafeCompare($configuredSecret, $key))
		{
			throw new SecretMismatch();
		}
	}
}