<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2023 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\Component\ATS\Administrator\Helper;

defined('_JEXEC') or die;

use Exception;
use Joomla\CMS\Factory;
use Joomla\CMS\Filter\InputFilter;
use Joomla\CMS\Log\Log;
use Joomla\CMS\Log\LogEntry;

/**
 * Handles logging and log management
 *
 * @since   5.0.0
 */
abstract class Logger
{
	private static $registeredLoggers = [];

	/**
	 * Register a CLI output logger for the given context
	 *
	 * If no callback is specified we will use the default callback which writes the log message to the output, either
	 * using the current CLI app's out() method or, if this is not available, a simple echo.
	 *
	 * The minimum log level to write to the file is determined by Joomla's debug flag. If you have enabled Site Debug
	 * the log level is JLog::All which log everything, including debug information. If Site Debug is disabled the
	 * log level is JLog::INFO which logs everything BUT debug information.
	 *
	 * @param   string         $context
	 * @param   callable|null  $callback
	 *
	 * @return  void
	 *
	 * @since   3.2.0
	 */
	public static function registerCLIOutputLogger(string $context, ?callable $callback = null): void
	{
		// Make sure we are not double-registering a logger
		$sig = md5($context . '.file');

		if (in_array($sig, self::$registeredLoggers))
		{
			return;
		}

		self::$registeredLoggers[] = $sig;

		// If no callback was provided we will use our default one
		if (empty($callback) || !is_callable($callback))
		{
			try
			{
				$app = Factory::getApplication();
			}
			catch (Exception $e)
			{
				$app = null;
			}

			if (is_object($app) && !method_exists($app, 'out'))
			{
				$app = null;
			}

			$callback = function (LogEntry $entry) use ($app) {
				if (is_null($app))
				{
					echo $entry->message . "\n";

					return;
				}

				$app->out($entry->message);
			};
		}

		// Register the CLI output logger
		$logLevel = Debug::getJoomlaDebug() ? Log::ALL : Log::INFO;

		Log::addLogger([
			'logger'   => 'callback',
			'callback' => $callback,
		], $logLevel, [$context]);

	}

	/**
	 * Register a file logger for the given context if we have not already done so.
	 *
	 * If no file is specified a log file will be created, named after the context. For example, the context 'ats.cron'
	 * is logged to the file 'ats_cron.php' in Joomla's configured logs directory.
	 *
	 * The minimum log level to write to the file is determined by Joomla's debug flag. If you have enabled Site Debug
	 * the log level is JLog::All which log everything, including debug information. If Site Debug is disabled the
	 * log level is JLog::INFO which logs everything BUT debug information.
	 *
	 * @param   string       $context
	 * @param   string|null  $file
	 *
	 * @return  void
	 *
	 * @since   3.2.0
	 */
	public static function registerFileLogger(string $context, ?string $file = null): void
	{
		// Make sure we are not double-registering a logger
		$sig = md5($context . '.file');

		if (in_array($sig, self::$registeredLoggers))
		{
			return;
		}

		self::$registeredLoggers[] = $sig;

		/**
		 * If no file is specified we will create a filename based on the context.
		 *
		 * For example the context 'ats.cron' results in the log filename 'ats_cron.php'
		 */
		if (is_null($file))
		{
			$filter          = InputFilter::getInstance();
			$filteredContext = $filter->clean($context, 'cmd');
			$file            = str_replace('.', '_', $filteredContext) . '.php';
		}

		// Register the file logger
		$logLevel = Debug::getJoomlaDebug() ? Log::ALL : Log::INFO;

		Log::addLogger(['text_file' => $file], $logLevel, [$context]);
	}
}