<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2023 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\Component\ATS\Administrator\Helper;

defined('_JEXEC') or die;

use Exception;
use Joomla\CMS\Factory;
use Joomla\CMS\Helper\TagsHelper;
use Joomla\CMS\Table\User;
use Joomla\Database\DatabaseDriver;

/**
 * A user tagging helper.
 *
 * Here's a fun fact. Joomla 4 ships with a #__content_types record for users BUT com_users does not support tagging or
 * versioning, neither does the core Joomla User table. This helper is a custom implementation of what would happen
 * under the hood should the core User table be taggable.
 *
 * @since  5.0.0
 */
abstract class UserTags
{
	/**
	 * UCM content type for users.
	 *
	 * @var   string
	 * @since 5.0.0
	 */
	private static $typeAlias = 'com_users.user';

	/**
	 * Get the tag IDs for a given user
	 *
	 * @param   int|null  $userId  The user ID to get tags for. NULL for current user.
	 *
	 * @return  int[]  The tag IDs
	 *
	 * @throws Exception
	 * @since  5.0.0
	 */
	public static function getUserTags(?int $userId = null): array
	{
		$userId = $userId ?? Permissions::getUser()->id;

		$tagsHelper            = new TagsHelper();
		$tagsHelper->typeAlias = self::$typeAlias;

		return explode(',', $tagsHelper->getTagIds($userId, self::$typeAlias));
	}

	/**
	 * Remove all tags from a user
	 *
	 * @param   int|null  $userId  The user ID to remove tags from. NULL for current user.
	 *
	 * @return  bool
	 *
	 * @throws Exception
	 * @since  5.0.0
	 */
	public static function removeUserTags(?int $userId = null): bool
	{
		$userId = $userId ?? Permissions::getUser()->id;

		$tagsHelper            = new TagsHelper();
		$tagsHelper->typeAlias = self::$typeAlias;

		/** @var DatabaseDriver $db */
		$db        = Factory::getContainer()->get('DatabaseDriver');
		$userTable = new User($db);

		if (!$userTable->load($userId))
		{
			return false;
		}

		return $tagsHelper->deleteTagData($userTable, $userId);
	}

	/**
	 * Set the tags for a user.
	 *
	 * New tags are marked as ##new##tagName in the $newTag list / array
	 *
	 * @param   int|null      $userId   The user ID to set the tags for. NULL for current user.
	 * @param   string|array  $newTags  Comma–separated string of tag IDs or array of tags IDs.
	 *
	 * @return  bool
	 *
	 * @throws  Exception
	 * @since   5.0.0
	 */
	public static function setUserTags(?int $userId = null, $newTags = ''): bool
	{
		$userId = $userId ?? Permissions::getUser()->id;

		$tagsHelper            = new TagsHelper();
		$tagsHelper->typeAlias = self::$typeAlias;

		/** @var DatabaseDriver $db */
		$db        = Factory::getContainer()->get('DatabaseDriver');
		$userTable = new User($db);

		if (!$userTable->load($userId))
		{
			return false;
		}

		if (empty($newTags))
		{
			$tagsHelper->preStoreProcess($userTable);

			return $tagsHelper->postStoreProcess($userTable);
		}

		$tagsHelper->preStoreProcess($userTable, (array) $newTags);

		if (is_string($newTags) && (strpos($newTags, ',') !== false))
		{
			$newTags = explode(',', $newTags);
		}
		elseif (!is_array($newTags))
		{
			$newTags = (array) $newTags;
		}

		return $tagsHelper->postStoreProcess($userTable, $newTags);
	}

}