<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2023 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\Component\ATS\Administrator\MailAuthentication;

defined('_JEXEC') or die;

use Joomla\CMS\Factory;
use Joomla\CMS\Http\HttpFactory;
use Joomla\CMS\Language\Text;
use Joomla\CMS\Uri\Uri;
use Joomla\Http\Response;
use RuntimeException;

class AbstractMailAuthentication implements MailAuthenticationInterface
{
	/**
	 * The internal email service name
	 *
	 * @since 5.0.0
	 * @var   string
	 */
	protected $name = '';

	/**
	 * The OAuth2 response type when asking for a code (which will be exchanged for a token)
	 *
	 * @since 5.2.0
	 * @var   string
	 */
	protected $responseType = 'code';

	/**
	 * The separator to use between multiple scopes
	 *
	 * @since 5.0.0
	 * @var   string
	 */
	protected $scopeSeparator = ' ';

	/**
	 * The scopes to request from the OAuth2 service
	 *
	 * @since 5.0.0
	 * @var   string[]
	 */
	protected $scopes = [];

	/**
	 * Service OAuth2 consent URL (without the question mark and any data
	 *
	 * @since 5.0.0
	 * @var   string
	 */
	protected $urlConsent;

	/**
	 * Service URL to exchange the OAuth2 code with access and refresh tokens
	 *
	 * @since 5.0.0
	 * @var   string
	 */
	protected $urlToken;

	/**
	 * @inheritDoc
	 */
	public function getRedirectionUrl(string $clientId): string
	{
		return $this->urlConsent . '?' . http_build_query([
				'client_id'     => $clientId,
				'redirect_uri'  => $this->getCallbackUrl(),
				'scope'         => implode($this->scopeSeparator, $this->scopes),
				'access_type'   => 'offline',
				'prompt'        => 'consent',
				'response_type' => $this->responseType,
				'state'         => Factory::getApplication()->getFormToken(),
			]);
	}

	/**
	 * @inheritDoc
	 */
	public function getTokens(string $code, string $clientId, string $secret): array
	{
		try
		{
			$response = $this->exchangeCodeForToken($code, $clientId, $secret);
		}
		catch (RuntimeException $e)
		{
			throw new ServiceErrorException('generic', $e->getMessage());
		}

		if ($response->code != 200)
		{
			$body = $response->body ?? '';

			try
			{
				$jsonDecoded = @json_decode($body, true);
			}
			catch (\Exception $e)
			{
				$jsonDecoded = null;
			}

			if (empty($body) || empty($jsonDecoded))
			{
				throw new ServiceErrorException('generic', Text::sprintf('COM_ATS_AUTHENTICATE_ERR_INVALID_HTTP_STATUS', $response->code));
			}
		}

		$result = @json_decode($response->body ?? '', true);

		if (is_null($result))
		{
			throw new ServiceErrorException('invalid_json', Text::_('COM_ATS_AUTHENTICATE_ERR_INVALID_JSON'));
		}

		// Do we have an error reported by Google Mail?
		if (isset($result['error']))
		{
			throw new ServiceErrorException($result['error'] ?? 'generic', $result['error_description'] ?? '', $result['error_uri'] ?? null);
		}

		return [
			'access'  => $result['access_token'] ?? '',
			'refresh' => $result['refresh_token'] ?? null,
		];
	}

	/**
	 * Exchanges an OAuth2 code for a token pair
	 *
	 * @param   string  $code      The code to exchange
	 * @param   string  $clientId  The OAuth2 application's client ID
	 * @param   string  $secret    The OAuth2 application's secret
	 *
	 * @return  Response
	 *
	 * @since   5.2.0
	 */
	protected function exchangeCodeForToken(string $code, string $clientId, string $secret): Response
	{
		return HttpFactory::getHttp()->post($this->urlToken, [
			'code'          => $code,
			'client_id'     => $clientId,
			'client_secret' => $secret,
			'grant_type'    => 'authorization_code',
			'redirect_uri'  => $this->getCallbackUrl(),
		]);
	}

	/**
	 * Get the callback URL for OAuth2 operations
	 *
	 * @return  string
	 *
	 * @since   5.0.0
	 */
	protected function getCallbackUrl(): string
	{
		return rtrim(Uri::base(), '/') . sprintf(
				'/index.php?option=com_ats&view=authenticate&task=callback&service=%s&tmpl=component',
				$this->name
			);
	}
}