<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2023 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\Component\ATS\Administrator\Model;

defined('_JEXEC') or die;

use Akeeba\Component\ATS\Administrator\Helper\Permissions;
use Akeeba\Component\ATS\Administrator\Mixin\ModelGetItemTrait;
use Akeeba\Component\ATS\Administrator\Table\ManagernoteTable;
use Akeeba\Component\ATS\Administrator\Table\TicketTable;
use Exception;
use Joomla\CMS\Application\CMSApplication;
use Joomla\CMS\Factory;
use Joomla\CMS\Form\Form;
use Joomla\CMS\MVC\Model\AdminModel;
use Joomla\CMS\Object\CMSObject;

#[\AllowDynamicProperties]
class ManagernoteModel extends AdminModel
{
	use ModelGetItemTrait;

	/**
	 * @inheritDoc
	 */
	public function getForm($data = [], $loadData = true)
	{
		$id       = $data['id'] ?? null;
		$formName = empty($id) ? 'managernote_new' : 'managernote';

		$form = $this->loadForm(
			'com_ats.' . $formName,
			$formName,
			[
				'control'   => 'jform',
				'load_data' => $loadData,
			]
		) ?: false;

		if (empty($form))
		{
			return false;
		}

		// Modify the form based on access controls.
		$ticketId = $data['ticket_id'] ?? $form->getValue('ticket_id');
		/** @var TicketTable $ticket */
		$ticket = $this->getTable('Ticket', 'Administrator');
		$ticket->load($ticketId);
		$canDo   = Permissions::getTicketPrivileges($ticket);
		$isOwner = ($data['created_by'] ?? -1) == Permissions::getUser()->id;

		$canEditState = $canDo['admin'] || $canDo['edit.state'] || $canDo['notes.edit'] || ($canDo['notes.edit.own'] && $isOwner);
		$canCreate    = $canDo['admin'] || $canDo['notes.create'];
		$canEdit      = $canDo['admin'] || $canDo['notes.edit'] || ($canDo['notes.edit.own'] && $isOwner);

		$fieldChanges = [
			'enabled'     => $canEditState ? '' : 'disable',
			'created'     => $canEditState ? '' : 'remove',
			'created_by'  => $canEditState ? '' : 'remove',
			'modified'    => $canEditState ? '' : 'remove',
			'modified_by' => $canEditState ? '' : 'remove',
			'note_html'   => ($canCreate || $canEdit) ? '' : 'remove',
		];

		array_walk($fieldChanges, function (string $type, string $fieldName, Form $form) {
			switch ($type)
			{
				case 'disable':
					$form->setFieldAttribute($fieldName, 'disabled', 'true');
					$form->setFieldAttribute($fieldName, 'required', 'false');
					$form->setFieldAttribute($fieldName, 'filter', 'unset');
					break;

				case 'remove':
					$form->removeField($fieldName);
					break;
			}
		}, $form);

		/**
		 * self::loadFormData() does not take $data into account. However, we use that to communicate overrides to the
		 * reply form, including the ticket ID. We need to make sure the $data array is taken into account indeed.
		 */
		if ($loadData)
		{
			$form->bind(array_merge($form->getData()->toArray(), $data));
		}

		return $form;
	}

	/**
	 * Method to get a single record.
	 *
	 * @param   int|null  $pk  The id of the primary key.
	 *
	 * @return  ManagernoteTable|bool  Object on success, false on failure.
	 * @throws  Exception
	 * @since        5.0.0
	 *
	 * @noinspection PhpMissingParentCallCommonInspection
	 */
	public function getItem($pk = null)
	{
		return $this->getItemTable($pk);
	}

	/** @inheritdoc */
	protected function canDelete($record)
	{
		/** @var TicketTable $ticket */
		$ticket = $this->getTable('Ticket', 'Administrator');

		if (!empty($record) && isset($record->ticket_id) && empty($record->ticket_id))
		{
			$ticket->load($record->ticket_id);
		}

		$canDo     = Permissions::getTicketPrivileges($ticket);

		if ($canDo['admin'])
		{
			return true;
		}

		if (!$canDo['notes.read'])
		{
			return false;
		}

		return $canDo['delete'] || $canDo['notes.delete'];
	}

	/** @inheritdoc */
	protected function canEditState($record)
	{
		/** @var TicketTable $ticket */
		$ticket = $this->getTable('Ticket', 'Administrator');

		if (!empty($record) && isset($record->ticket_id) && empty($record->ticket_id))
		{
			$ticket->load($record->ticket_id);
		}

		$canDo     = Permissions::getTicketPrivileges($ticket);

		if ($canDo['admin'])
		{
			return true;
		}

		if (!$canDo['notes.read'])
		{
			return false;
		}

		if ($canDo['edit.state'] || $canDo['notes.edit'])
		{
			return true;
		}

		if ($record->created_by != Permissions::getUser()->id)
		{
			return false;
		}

		return $canDo['notes.edit.own'];
	}

	/**
	 * Load the data of an add / edit form.
	 *
	 * The data is loaded from the user state. If the user state is empty we load the item being edited. If there is no
	 * item being edited we will override the default table values with the respective list filter values. This makes
	 * sense for users. If I am filtering by category X and maturity Stable I am probably trying to see if there is a
	 * specific stable version released in category X and, if not, create it. Using the filter values reduces the
	 * possibility for silly mistakes on the part of the operator.
	 *
	 * @return array|bool|CMSObject|mixed
	 * @throws Exception
	 */
	protected function loadFormData()
	{
		/** @var CMSApplication $app */
		$app  = Factory::getApplication();
		$data = $app->getUserState('com_ats.edit.managernote.data', []);

		if (empty($data))
		{
			$data = $this->getItem();
		}

		$this->preprocessData('com_ats.managernote', $data);

		return $data;
	}

	/** @inheritdoc */
	protected function prepareTable($table)
	{
		// Set up the created / modified date
		$date  = Factory::getDate();
		$user  = Factory::getApplication()->getIdentity();
		$isNew = empty($table->getId());

		if ($isNew)
		{
			// Set the values
			$table->created    = $date->toSql();
			$table->created_by = $user->id;
		}
		else
		{
			// Set the values
			$table->modified    = $date->toSql();
			$table->modified_by = $user->id;
		}
	}


}