<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2023 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\Component\ATS\Administrator\Model;

defined('_JEXEC') or die;

use Akeeba\Component\ATS\Administrator\Mixin\ComponentObjectTrait;
use Akeeba\Component\ATS\Administrator\Service\CacheCleaner;
use Joomla\CMS\Application\CMSApplication;
use Joomla\CMS\Component\ComponentHelper;
use Joomla\CMS\Factory;
use Joomla\CMS\Language\Text;
use Joomla\CMS\MVC\Model\BaseDatabaseModel;
use Joomla\CMS\Plugin\PluginHelper;
use Joomla\CMS\Uri\Uri;
use Joomla\CMS\User\User;
use Joomla\CMS\User\UserHelper;
use Joomla\Filter\InputFilter;
use Joomla\Registry\Registry;
use RuntimeException;

#[\AllowDynamicProperties]
class SchedulingModel extends BaseDatabaseModel
{
	use ComponentObjectTrait;

	/**
	 * Cache of all enabled ATS CRON commands
	 *
	 * @since 3.2.0
	 * @var   array|null
	 */
	protected $atsCronCommands = null;

	/**
	 * Cache of all of the installed ATS CRON plugins
	 *
	 * @since 3.2.0
	 * @var   array|null
	 */
	protected $atsCronPlugins = null;

	/**
	 * Apply the given secret. If none is provided the already generated, random Secret Word will be used instead.
	 *
	 * @param   string  $secret  The secret word to apply. Empty to use the already generated, random Secret Word.
	 *
	 * @return  void
	 *
	 * @since   3.2.0
	 */
	public function applySecretWord($secret = '')
	{
		/** @var CMSApplication $app */
		$app     = Factory::getApplication();
		$session = $app->getSession();
		$secret  = trim($secret);


		if (empty($secret))
		{
			$secret = $session->get('ats.cron.newSecretWord', '');
		}

		if (empty($secret))
		{
			$this->createNewSecretWord();

			$secret = $session->get('ats.cron.newSecretWord', '');
		}

		$cParams = ComponentHelper::getParams('com_ats');
		$cParams->set('secret', $secret);

		$this->getExtensionObject()
			->getComponentParametersService()
			->save($cParams);
	}

	/**
	 * Generate a new, random Secret Word and store it in the session
	 *
	 * @return  void
	 *
	 * @since   3.2.0
	 */
	public function createNewSecretWord()
	{
		/** @var CMSApplication $app */
		$app = Factory::getApplication();
		$app->getSession()->set('ats.cron.newSecretWord', UserHelper::genRandomPassword(32));
	}

	/**
	 * Return the information about the currently enabled ATS CRON commands
	 *
	 * @return  array
	 *
	 * @since   3.2.0
	 */
	public function getActiveCRONCommands()
	{
		if (!is_null($this->atsCronCommands))
		{
			return $this->atsCronCommands;
		}

		PluginHelper::importPlugin('ats');

		$responses = Factory::getApplication()->triggerEvent('onAtsCronTaskInfo', []);

		$this->atsCronCommands = array_filter($responses, function ($response) {
			if (empty($response) || !is_array($response))
			{
				return false;
			}

			foreach (['command', 'label', 'description'] as $key)
			{
				if (!array_key_exists($key, $response))
				{
					return false;
				}

				if (empty($response[$key]))
				{
					return false;
				}
			}

			return true;
		});

		$keys = array_map(function ($info) {
			return $info['command'];
		}, $this->atsCronCommands);

		$this->atsCronCommands = array_combine($keys, $this->atsCronCommands);

		return $this->atsCronCommands;
	}

	/**
	 * Get a list of all of the installed ATS CRON plugins
	 *
	 * @return  array
	 *
	 * @since   3.2.0
	 */
	public function getAtsCronPlugins()
	{
		// Return the cached results if available
		if (!is_null($this->atsCronPlugins))
		{
			return $this->atsCronPlugins;
		}

		// Default: no plugins detected
		$this->atsCronPlugins = [];

		// Get the information for all installed ATS plugins
		try
		{
			$pluginsInfo = $this->getAtsPluginsFromDB();

			if (is_null($pluginsInfo))
			{
				throw new RuntimeException('No ats plugins are installed on this site');
			}
		}
		catch (RuntimeException $e)
		{
			return $this->atsCronPlugins;
		}

		// Filter by plugins with an onAtsCronTaskInfo method
		$pluginsInfo = array_filter($pluginsInfo, function (array $pluginInfo) {
			$pluginPath     = JPATH_PLUGINS . '/ats/' . $pluginInfo['element'] . '/' . $pluginInfo['element'] . '.php';
			$pluginClass    = 'plgAts' . ucfirst($pluginInfo['element']);
			$pluginClassAlt = '\\Joomla\\Plugin\\Ats\\' . ucfirst($pluginInfo['element']) . '\\Extension\\' . ucfirst($pluginInfo['element']);

			// Does the plugin even exist?
			if (!@file_exists($pluginPath))
			{
				return false;
			}

			// If the plugin is already loaded I can check if the method exists directly
			if (class_exists($pluginClass))
			{
				return method_exists($pluginClass, 'onAtsCronTaskInfo');
			}

			if (class_exists($pluginClassAlt))
			{
				return method_exists($pluginClassAlt, 'onAtsCronTaskInfo');
			}

			$fileContents = @file_get_contents($pluginPath);

			if ($fileContents === false)
			{
				return false;
			}

			return preg_match('/function(\s){1,}onAtsCronTaskInfo/', $fileContents) > 0;
		});

		if (empty($pluginsInfo))
		{
			return $this->atsCronPlugins;
		}

		// Process the plugin info into something usable
		$this->atsCronPlugins = array_map(function ($info) {
			$params           = new Registry($info['params']);
			$manifest         = new Registry($info['manifest_cache']);
			$cronValue        = $params->get('cron', null);
			$user             = new User();
			$publicViewLevels = $user->getAuthorisedViewLevels();

			Factory::getApplication()->getLanguage()->load('plg_ats_' . $info['element'] . '.sys', JPATH_ADMINISTRATOR, null, false, true);

			$ret = [
				'id'          => $info['extension_id'],
				'name'        => Text::_($info['name']),
				'command'     => $info['element'],
				'description' => Text::_($manifest->get('description', '')),
				'published'   => $info['enabled'] == 1,
				'public'      => in_array($info['access'], $publicViewLevels),
				'cronEnabled' => is_null($cronValue) || ($cronValue == 1),
			];

			return $ret;
		}, $pluginsInfo);

		return $this->atsCronPlugins;
	}

	/**
	 * Get the CRON information which is used to render the Automation page
	 *
	 * @return  object
	 * @since   3.2.0
	 */
	public function getCRONInfo()
	{
		$ret = (object) [
			'cli'      => (object) [
				'supported' => false,
				'path'      => false,
			],
			'frontend' => (object) [
				'supported' => false,
				'path'      => false,
			],
			'info'     => (object) [
				'windows'  => false,
				'php_path' => false,
				'root_url' => false,
				'secret'   => '',
			],
		];

		$siteRoot = rtrim(realpath(JPATH_ROOT), '/\\');
		$cParams  = ComponentHelper::getParams('com_ats');

		$ret->info->windows  = (DIRECTORY_SEPARATOR == '\\') || (substr(strtoupper(PHP_OS), 0, 3) == 'WIN');
		$ret->info->php_path = $ret->info->windows ? 'c:\path\to\php.exe' : '/path/to/php';
		$ret->info->root_url = rtrim(Uri::base(), '/');
		$ret->info->secret   = $cParams->get('secret', '');

		if (substr($ret->info->root_url, -14) == '/administrator')
		{
			$ret->info->root_url = substr($ret->info->root_url, 0, -14);
		}

		$commandName = $this->getSelectedCRONCommand();

		// Get information for CLI CRON script
		$ret->cli->supported = true;
		$ret->cli->path      = implode(DIRECTORY_SEPARATOR, [$siteRoot, 'cli', 'joomla.php']);
		$ret->cli->path      .= ' ats:cron ' . $commandName;

		// Get information for front-end backup
		if (trim($ret->info->secret))
		{
			$ret->frontend->supported = true;
			$ret->frontend->path      = 'index.php?option=com_ats&view=Cron&key=' . urlencode($ret->info->secret) .
				'&command=' . urlencode($commandName);
		}

		return $ret;
	}

	/**
	 * Get the currently selected CRON command
	 *
	 * @return  string
	 *
	 * @since   3.2.0
	 */
	public function getSelectedCRONCommand()
	{
		/** @var CMSApplication $app */
		$app     = Factory::getApplication();
		$session = $app->getSession();

		$commandName = $session->get('ats.cron.commandName', 'commandName');
		$allCommands = $this->getActiveCRONCommands();

		// Check for validity
		if (!array_key_exists($commandName, $allCommands))
		{
			$commandName = 'commandName';
		}

		return $commandName;
	}

	/**
	 * Has the user configured a Secret Word for CRON jobs?
	 *
	 * @return  bool
	 *
	 * @since   3.2.0
	 */
	public function hasSecretWord()
	{
		$secret = trim(ComponentHelper::getParams('com_ats')->get('secret', ''));

		return !empty($secret);
	}

	/**
	 * Publish an ATS plugin with public access (or set the access to public if it's already published)
	 *
	 * @param   string  $pluginName  The name of the ats plugin
	 *
	 * @return  bool  True if the operation succeeded or the plugin was already published with a public access level.
	 *
	 * @since   3.2.0
	 */
	public function publishPlugin($pluginName)
	{
		// Get the filtered plugin name
		$filter     = new InputFilter();
		$pluginName = $filter->clean($pluginName, 'cmd');

		if (empty($pluginName) || is_array($pluginName))
		{
			return false;
		}

		// Make sure the plugin in question exists
		$atsCronPlugins = $this->getAtsCronPlugins();

		if (!array_key_exists($pluginName, $atsCronPlugins))
		{
			return false;
		}

		// Make sure the plugin in question is not already published with a public access level and CRON enabled
		$pluginInfo = $atsCronPlugins[$pluginName];
		$pluginId   = $pluginInfo['id'];

		if ($pluginInfo['published'] && $pluginInfo['public'] && $pluginInfo['cronEnabled'])
		{
			return true;
		}

		$db      = $this->getDatabase();
		$changes = [];

		// Do I need to publish the plugin?
		if (!$pluginInfo['published'])
		{
			$changes[] = $db->quoteName('enabled') . ' = 1';
		}

		// Do I need to make the plugin public?
		if (!$pluginInfo['public'])
		{
			// Get the first public view access level
			$user             = new User();
			$publicViewLevels = $user->getAuthorisedViewLevels();

			if (empty($publicViewLevels))
			{
				return false;
			}

			$publicLevel = min($publicViewLevels);

			if (empty($publicLevel))
			{
				return false;
			}

			$changes[] = $db->qn('access') . ' = ' . $db->quote($publicLevel);
		}

		try
		{
			if (!empty($changes))
			{
				$query = $db->getQuery(true)
				            ->update($db->qn('#__extensions'))
				            ->set($changes)
				            ->where($db->qn('extension_id') . ' = :pluginId')
				            ->bind(':pluginId', $pluginId);;

				$db->setQuery($query)->execute();
			}

			$retResult = true;

			// Do I need to Enable CRON for this plugin?
			if (!$pluginInfo['cronEnabled'])
			{
				$pluginInfo = PluginHelper::getPlugin('ats', $pluginName);
				$params     = new Registry($pluginInfo->params);
				$params->set('cron', 1);

				$this->getExtensionObject()
				     ->getComponentParametersService()
				     ->save($params, sprintf("plg_ats_%s", $pluginName));
			}
		}
		catch (RuntimeException $e)
		{
			$retResult = false;
		}
		finally
		{
			/** @var CacheCleaner $cacheCleanerService */
			$cacheCleanerService = $this->bootComponent('com_ats')
			                            ->getCacheCleanerService();

			$cacheCleanerService->clearGroups(['com_plugins']);

			return $retResult;
		}
	}

	/**
	 * Get a list of installed ATS plugins from the database
	 *
	 * @return  array|null
	 *
	 * @since   3.2.0
	 */
	protected function getAtsPluginsFromDB()
	{
		$db = $this->getDatabase();

		$query = $db->getQuery(true)
		            ->select([
			            $db->quoteName('extension_id'),
			            $db->quoteName('name'),
			            $db->quoteName('element'),
			            $db->quoteName('enabled'),
			            $db->quoteName('access'),
			            $db->quoteName('params'),
			            $db->quoteName('manifest_cache'),
		            ])
		            ->from($db->quoteName('#__extensions'))
		            ->where($db->quoteName('type') . ' = ' . $db->q('plugin'))
		            ->where($db->quoteName('folder') . ' = ' . $db->q('ats'));

		return $db->setQuery($query)->loadAssocList('element');
	}
}