<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2023 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\Component\ATS\Administrator\View\Cannedreplies;

use Akeeba\Component\ATS\Administrator\Mixin\ViewLoadAnyTemplateTrait;
use Akeeba\Component\ATS\Administrator\Mixin\ViewTaskBasedEventsTrait;
use Akeeba\Component\ATS\Administrator\Model\CannedrepliesModel;
use Akeeba\Component\ATS\Administrator\Table\CannedreplyTable;
use Exception;
use Joomla\CMS\Factory;
use Joomla\CMS\Form\Form;
use Joomla\CMS\Language\Text;
use Joomla\CMS\MVC\View\GenericDataException;
use Joomla\CMS\MVC\View\HtmlView as BaseHtmlView;
use Joomla\CMS\Pagination\Pagination;
use Joomla\CMS\Router\Route;
use Joomla\CMS\Toolbar\Button\DropdownButton;
use Joomla\CMS\Toolbar\Toolbar;
use Joomla\CMS\Toolbar\ToolbarHelper;
use Joomla\CMS\Uri\Uri;
use Joomla\Registry\Registry;

defined('_JEXEC') or die;

class HtmlView extends BaseHtmlView
{
	use ViewLoadAnyTemplateTrait;
	use ViewTaskBasedEventsTrait;

	/**
	 * The active search filters
	 *
	 * @var    array
	 * @since  5.0.0
	 */
	public $activeFilters = [];

	/**
	 * The name of the editor in the parent window of this modal
	 *
	 * @var   string|null
	 * @since 5.0.0
	 */
	public $editor;

	/**
	 * The search tools form
	 *
	 * @var    Form
	 * @since  5.0.0
	 */
	public $filterForm;

	/**
	 * An array of items
	 *
	 * @var    CannedreplyTable[]
	 * @since  5.0.0
	 */
	protected $items = [];

	/**
	 * The pagination object
	 *
	 * @var    Pagination
	 * @since  5.0.0
	 */
	protected $pagination;

	/**
	 * The model state
	 *
	 * @var    Registry
	 * @since  5.0.0
	 */
	protected $state;

	/**
	 * Is this view an Empty State
	 *
	 * @var   bool
	 * @since 5.0.0
	 */
	private $isEmptyState = false;

	public function display($tpl = null)
	{
		/** @var CannedrepliesModel $model */
		$model               = $this->getModel();
		$this->items         = $model->getItems();
		$this->pagination    = $model->getPagination();
		$this->state         = $model->getState();
		$this->filterForm    = $model->getFilterForm();
		$this->activeFilters = $model->getActiveFilters();

		// Check for errors.
		if (count($errors = $this->get('Errors')))
		{
			throw new GenericDataException(implode("\n", $errors), 500);
		}

		if (!\count($this->items) && $this->isEmptyState = $this->get('IsEmptyState'))
		{
			$this->setLayout('emptystate');
		}

		if (!empty($this->editor))
		{
			$this->document->addScriptOptions('atsParentEditor', $this->editor);
		}
		else
		{
			$this->addToolbar();
		}

		parent::display($tpl);
	}

	/**
	 * Rebase (most) links and images so they work correctly in the backend.
	 *
	 * @param   string|null  $content
	 *
	 * @return  string
	 *
	 * @throws  Exception
	 * @since   5.0.0
	 */
	public function rebaseLinks(string $content): string
	{
		static $appPrefix = null;

		if (is_null($appPrefix))
		{
			$app = Factory::getApplication();
			$appPrefix = '';

			if ($app->isClient('administrator') || $app->isClient('api'))
			{
				$appPrefix = $app->getName() . '/';
			}
		}

		if (empty($content))
		{
			return '';
		}

		if (empty($appPrefix))
		{
			return $content;
		}

		return preg_replace_callback('#(href|src)="([^"]*)"#im', function ($matches) use ($appPrefix) {
			$attribute = $matches[1];
			$link      = $matches[2];

			$uri  = Uri::getInstance($link);
			$base = $uri->toString(['scheme', 'host', 'port', 'path']);
			$host = $uri->toString(['scheme', 'host', 'port']);
			$siteBase = Uri::base();

			if (!empty($appPrefix) && substr($siteBase, -strlen($appPrefix)) === $appPrefix)
			{
				$siteBase = substr($siteBase, 0, -strlen($appPrefix));
			}

			$isInternalish = empty($host) ||
				preg_match('#' . preg_quote($siteBase, '#') . '#', $base) ||
				($host === Uri::getInstance($siteBase)->getHost() && strpos($uri->getPath(), 'index.php') !== false);

			if ($isInternalish)
			{
				$canonical = Uri::getInstance(Uri::base());
				$uri->setScheme($canonical->getScheme());
				$uri->setHost($canonical->getHost());
				$uri->setPort($canonical->getPort());

				$path = ltrim($uri->getPath(), '/');

				if (strpos($path, $appPrefix) === 0)
				{
					$path = substr($path, strlen($appPrefix));
				}

				$uri->setPath('/' . ltrim($path, '/'));

				$link = $uri;
			}

			return sprintf('%s="%s"', $attribute, $link);
		}, $content);
	}

	private function addToolbar()
	{
		$user = Factory::getApplication()->getIdentity();

		// Get the toolbar object instance
		$toolbar = Toolbar::getInstance('toolbar');

		ToolbarHelper::title(sprintf(Text::_('COM_ATS_TITLE_CANNEDREPLIES')), 'icon-ats');

		$canCreate    = $user->authorise('core.create', 'com_ats');
		$canDelete    = $user->authorise('core.delete', 'com_ats');
		$canEditState = $user->authorise('core.edit.state', 'com_ats');

		if ($canCreate)
		{
			$toolbar
				->addNew('cannedreply.add');
		}

		if ($canDelete || $canEditState)
		{
			/** @var DropdownButton $dropdown */
			$dropdown = $toolbar->dropdownButton('status-group')
				->text('JTOOLBAR_CHANGE_STATUS')
				->toggleSplit(false)
				->icon('icon-ellipsis-h')
				->buttonClass('btn btn-action')
				->listCheck(true);

			$childBar = $dropdown->getChildToolbar();

			if ($canEditState)
			{
				$childBar->publish('cannedreplies.publish')
					->icon('fa fa-check-circle')
					->text('JTOOLBAR_PUBLISH')
					->listCheck(true);

				$childBar->unpublish('cannedreplies.unpublish')
					->icon('fa fa-times-circle')
					->text('JTOOLBAR_UNPUBLISH')
					->listCheck(true);

				$childBar->checkin('cannedreplies.checkin')->listCheck(true);
			}

			if ($canDelete)
			{
				$childBar->delete('cannedreplies.delete')
					->message('JGLOBAL_CONFIRM_DELETE')
					->listCheck(true);
			}
		}

		ToolbarHelper::back('JTOOLBAR_BACK', Route::_('index.php?option=com_ats&view=Controlpanel', false));
	}

}