<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2023 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\Component\ATS\Site\Controller;

defined('_JEXEC') or die;

use Akeeba\Component\ATS\Administrator\Exceptions\CronCommandMissing;
use Akeeba\Component\ATS\Administrator\Exceptions\CronCommandNotFound;
use Akeeba\Component\ATS\Administrator\Exceptions\SecretMismatch;
use Akeeba\Component\ATS\Administrator\Exceptions\SecretNotConfigured;
use Akeeba\Component\ATS\Administrator\Helper\Cron;
use Exception;
use Joomla\CMS\Application\CMSApplication;
use Joomla\CMS\Component\ComponentHelper;
use Joomla\CMS\Factory;
use Joomla\CMS\Log\Log;
use Joomla\CMS\MVC\Controller\BaseController;
use Throwable;

class CronController extends BaseController
{
	/**
	 * Main and only entry point: executes a CRON command
	 *
	 * This is meant to be accessed as:
	 * index.php?option=com_ats&view=cron&key=SECRET_KEY&command=CRON_COMMAND
	 *
	 * @return  void  This never returns; it's an immediate exit point.
	 *
	 * @since   3.2.0
	 */
	public function display($cachable = false, $urlparams = [])
	{
		Cron::log('Starting the CRON job from URL');

		// Get a reference to the application or die trying
		try
		{
			/** @var CMSApplication $app */
			$app = Factory::getApplication();
		}
		catch (Throwable $e)
		{
			Cron::log($e->getMessage(), Log::ERROR);
			Cron::log('Exit: Cannot instantiate application');

			$this->bellyUp('Internal Server Error');

			// Technically unnecessary but helps with static code analysis :)
			return;
		}

		// Check the secret key
		$key = $this->input->get->get('key', '', 'raw');

		Cron::log('Checking secret key');

		try
		{
			Cron::checkSecret($key);
		}
		catch (SecretMismatch $e)
		{
			Cron::log('Provided secret key does not match configuration', Log::ERROR);
			Cron::log('Exit: secret key mismatch');

			$this->bellyUp('Forbidden', $e->getCode());

			// Technically unnecessary but helps with static code analysis :)
			return;
		}
		catch (SecretNotConfigured $e)
		{
			Cron::log('A secret key has not been configured yet', Log::ERROR);
			Cron::log('Exit: secret key not yet configured');

			$this->bellyUp('Service Unavailable', $e->getCode());

			// Technically unnecessary but helps with static code analysis :)
			return;
		}

		$command = $this->input->get->getCmd('command', '');

		// You might ask for a 'cmd' filter but get an array instead
		if (!is_string($command))
		{
			Cron::log('The command specified in the URL has an invalid format', Log::ERROR);
			Cron::log('Exit: invalid command format');

			$this->bellyUp('Bad Request', 400);

			// Technically unnecessary but helps with static code analysis :)
			return;
		}

		Cron::log('Executing the command');

		// Run the command
		try
		{
			$cParams   = ComponentHelper::getParams('com_ats');
			$timeLimit = (int) $cParams->get('time_limit', 10);
			$success   = Cron::run($command, $timeLimit);
		}
		catch (CronCommandMissing $e)
		{
			Cron::log('There was no command specified in the URL', Log::ERROR);
			Cron::log('Exit: no command');

			$this->bellyUp('Bad Request', 400);

			// Technically unnecessary but helps with static code analysis :)
			return;
		}
		catch (CronCommandNotFound $e)
		{
			Cron::log(sprintf('The requested command, “%s”, is not implemented', $command), Log::ERROR);
			Cron::log('Exit: unknown command');

			$this->bellyUp('Not Implemented', 501);

			// Technically unnecessary but helps with static code analysis :)
			return;
		}
		catch (Throwable $e)
		{
			Cron::log(sprintf('An error occurred executing command “%s”', $command), Log::ERROR);
			Cron::log($e->getMessage());
			Cron::log($e->getFile() . '::' . $e->getLine());

			foreach (explode("\n", $e->getTraceAsString()) as $line)
			{
				Cron::log($line);
			}

			Cron::log('Exit: error executing command');
			$this->bellyUp('Internal Server Error');

			// Technically unnecessary but helps with static code analysis :)
			return;
		}

		if ($success)
		{
			@header('Expires: Wed, 17 Aug 2005 00:00:00 GMT', true, 200);
			@header('Last-Modified: ' . gmdate('D, d M Y H:i:s') . ' GMT', true);
			@header('X-Cache-Control: False', true);
			@header('Cache-Control: no-store, no-cache, must-revalidate, post-check=0, pre-check=0', true);
			@header('Pragma: no-cache', true);

			Cron::log('Exit: COMMAND SUCCEEDED');

			echo "OK";

			$app->close();

			// Technically unnecessary but helps with static code analysis :)
			return;
		}

		Cron::log('Exit: COMMAND FAILED');

		$this->bellyUp('Command failed');
	}

	/**
	 * Exit the script with an error
	 *
	 * @param   string  $reason    The reason we're quitting. Will be included in the HTTP header.
	 * @param   int     $httpCode  HTTP status code, default 500
	 *
	 * @return  void  We actually never return; this is an exit point.
	 * @since   3.2.0
	 */
	private function bellyUp(string $reason, int $httpCode = 500): void
	{
		@header(sprintf('HTTP/1.1 %u %s', $httpCode, $reason), true, $httpCode);
		@header('Expires: Wed, 17 Aug 2005 00:00:00 GMT', true);
		@header('Last-Modified: ' . gmdate('D, d M Y H:i:s') . ' GMT', true);
		@header('X-Cache-Control: False', true);
		@header('Cache-Control: no-store, no-cache, must-revalidate, post-check=0, pre-check=0', true);
		@header('Pragma: no-cache', true);

		try
		{
			Factory::getApplication()->close();
		}
		catch (Throwable $e)
		{
			exit;
		}
	}
}