<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2023 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Joomla\Plugin\Ats\Autoclose\Extension;

defined('_JEXEC') or die;

use Akeeba\Component\ATS\Administrator\Helper\Timer;
use Akeeba\Component\ATS\Administrator\Mixin\ControllerNewPostTrait;
use Akeeba\Component\ATS\Administrator\Table\PostTable;
use Akeeba\Component\ATS\Administrator\Table\TicketTable;
use Exception;
use Joomla\CMS\Factory;
use Joomla\CMS\Language\Text;
use Joomla\CMS\Log\Log;
use Joomla\CMS\Plugin\CMSPlugin;
use Joomla\Database\DatabaseAwareInterface;
use Joomla\Database\DatabaseAwareTrait;
use Joomla\Database\DatabaseDriver;
use Joomla\Database\ParameterType;
use Joomla\Event\Event;
use Joomla\Event\SubscriberInterface;

/**
 * Automatically closes stale tickets
 *
 * @since 3.2.0
 */
class Autoclose extends CMSPlugin implements SubscriberInterface, DatabaseAwareInterface
{
	use ControllerNewPostTrait;
	use DatabaseAwareTrait;

	const atsCommandName = 'autoclose';

	/** @inheritdoc */
	protected $autoloadLanguage = true;

	public static function getSubscribedEvents(): array
	{
		return [
			'onAtsCronTaskInfo' => 'onAtsCronTaskInfo',
			'onAtsCronTask'     => 'onAtsCronTask',
		];
	}

	/**
	 * Handles the ATS CRON tasks.
	 *
	 * @param   Event  $event
	 *
	 * @return  void
	 * @since   3.2.0
	 */
	public function onAtsCronTask(Event $event): void
	{
		/**
		 * @var string $command The command name ATS CRON was asked to execute.
		 * @var array  $options Any options passed to this CRON job
		 */
		[$command, $options] = $event->getArguments();
		$result = $event->getArgument('result', []);

		// Make sure we are calling the correct command
		if ($command != self::atsCommandName)
		{
			return;
		}

		$event->setArgument('result', array_merge($result, [true]));

		$timeLimit  = array_key_exists('time_limit', $options) ? $options['time_limit'] : 86400;
		$timer      = new Timer($timeLimit);
		$max_age    = $this->params->get('maxinactiveticket', 30);
		$silent_age = $this->params->get('silentticketcloseage', 60);

		Log::add(sprintf("Closing tickets inactive over %s days", $max_age), Log::DEBUG, 'ats.cron');
		Log::add(sprintf("Silent close for tickets older than %s days", $silent_age), Log::DEBUG, 'ats.cron');

		// Load the language files
		$jlang = $this->getApplication()->getLanguage();
		$jlang->load('com_ats', JPATH_ADMINISTRATOR);
		$jlang->load('com_ats.override', JPATH_ADMINISTRATOR);

		// Loop while we have tickets and we have not hit the time limit
		while ($timer->getTimeLeft())
		{
			$noSilentLeft  = $silent_age < 1;
			$noVerboseLeft = false;

			if ($silent_age >= 1)
			{
				Log::add("Looking for up to 100 tickets to close silently", Log::DEBUG, 'ats.cron');

				$until   = time() - $silent_age * 3600 * 24;
				$tickets = $this->getTicketsUntil($until);

				$numTickets = count($tickets);
				Log::add(sprintf("Found %d tickets to close silently", $numTickets), Log::DEBUG, 'ats.cron');

				if ($numTickets === 0)
				{
					$noSilentLeft = true;
				}

				foreach ($tickets as $ticket)
				{
					if ($timer->getTimeLeft() <= 0)
					{
						Log::add(sprintf('I have reached the time limit of %u second(s). Stopping for now.', $timeLimit), Log::DEBUG, 'ats.cron');

						$event->setArgument('result', array_merge($event->getArgument('result', []), [true]));

						return;
					}

					Log::add(sprintf("\tSilently closing #%s %s", $ticket->getId(), $ticket->title), Log::DEBUG, 'ats.cron');

					TicketTable::$noPermissionsCheck = 1;
					$ticket->save([
						'status' => 'C',
					]);
					TicketTable::$noPermissionsCheck = 0;
				}
			}

			if ($timer->getTimeLeft() <= 0)
			{
				Log::add(sprintf('I have reached the time limit of %u second(s). Stopping for now.', $timeLimit), Log::DEBUG, 'ats.cron');

				$event->setArgument('result', array_merge($event->getArgument('result', []), [true]));

				return;
			}

			// Do I have to close tickets verbosely?
			if (($max_age < $silent_age) && ($max_age > 0))
			{
				Log::add("Looking for up to 100 tickets to close verbosely", Log::DEBUG, 'ats.cron');

				$until   = time() - $max_age * 3600 * 24;
				$tickets = $this->getTicketsUntil($until);

				$numTickets = count($tickets);
				Log::add(sprintf("Found %d tickets to close verbosely", $numTickets), Log::DEBUG, 'ats.cron');

				if ($numTickets == 0)
				{
					$noVerboseLeft = true;
				}

				$jNow = clone Factory::getDate();

				foreach ($tickets as $ticket)
				{
					if ($timer->getTimeLeft() <= 0)
					{
						Log::add(sprintf('I have reached the time limit of %u second(s). Stopping for now.', $timeLimit), Log::DEBUG, 'ats.cron');

						$event->setArgument('result', array_merge($event->getArgument('result', []), [true]));

						return;
					}

					Log::add(sprintf("\tVerbosely closing #%s %s", $ticket->ats_ticket_id, $ticket->title), Log::DEBUG, 'ats.cron');

					$post = [
						'ticket_id'    => $ticket->getId(),
						'content_html' => Text::_('COM_ATS_CLI_CLOSEPOSTMSG'),
						'created'      => $jNow->toSql(),
						'created_by'   => -1,
						'enabled'      => 1,
					];

					/** @var DatabaseDriver $database */
					$database  = $this->getDatabase();
					$postTable = new PostTable($database);

					$postTable->setUpdateCreated(false);
					$postTable->setUpdateModified(false);
					$postTable->save($post);
					$postTable->setUpdateCreated(true);
					$postTable->setUpdateModified(true);

					$this->postNotifiable($postTable, true);

					TicketTable::$noPermissionsCheck = 1;
					$ticket->save([
						'status' => 'C',
					]);
					TicketTable::$noPermissionsCheck = 0;
				}
			}
			else
			{
				$noVerboseLeft = true;
			}

			if ($noSilentLeft && $noVerboseLeft)
			{
				$event->setArgument('result', array_merge($event->getArgument('result', []), [true]));

				return;
			}
		}

		Log::add("Done auto-closing inactive tickets.", Log::DEBUG, 'ats.cron');

		$event->setArgument('result', array_merge($event->getArgument('result', []), [true]));
	}

	/**
	 * Returns information about the CRON task provided by this plugin
	 *
	 * @param   Event  $event
	 *
	 * @return  void
	 *
	 * @since   3.2.0
	 */
	public function onAtsCronTaskInfo(Event $event): void
	{
		$result = $event->getArgument('result', []);
		$event->setArgument('result', array_merge($result, [
			[
				'command'     => self::atsCommandName,
				'label'       => 'PLG_ATS_' . $this->_name . '_TASK_LABEL',
				'description' => 'PLG_ATS_' . $this->_name . '_TASK_DESC',
			],
		]));
	}

	/**
	 * Get a Tickets collection for up to 100 ATS tickets modified up to and including the $until date string.
	 *
	 * @param   string  $until
	 *
	 * @return  TicketTable[]
	 */
	private function getTicketsUntil(string $until): array
	{
		$jUntil    = clone Factory::getDate($until);
		$jUntilSql = $jUntil->toSql();

		/** @var DatabaseDriver $db */
		$db    = $this->getDatabase();
		$query = $db->getQuery(true)
			->select('*')
			->from($db->quoteName('#__ats_tickets'))
			->whereIn($db->quoteName('status'), ['O', 'P'], ParameterType::STRING)
			->where($db->quoteName('modified_on') . ' <= :until')
			->bind(':until', $jUntilSql);
		$query->setLimit(0, 100);
		$db->setQuery($query);

		$ticketTable = new TicketTable($db);

		try
		{
			return array_map(function ($x) use ($ticketTable) {
				$ticket = clone $ticketTable;
				$ticket->reset();
				$ticket->bind($x);

				return $ticket;
			}, $db->setQuery($query)->loadAssocList() ?: []);
		}
		catch (Exception $e)
		{
			return [];
		}
	}
}