<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2023 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Joomla\Plugin\Ats\Autoreply\Extension;

use Akeeba\Component\ATS\Administrator\Helper\AutoReply as AutoReplyHelper;
use Akeeba\Component\ATS\Administrator\Helper\Timer;
use Akeeba\Component\ATS\Administrator\Table\PostTable;
use Akeeba\Component\ATS\Administrator\Table\TicketTable;
use Joomla\CMS\Application\CMSApplication;
use Joomla\CMS\Log\Log;
use Joomla\CMS\Plugin\CMSPlugin;
use Joomla\Database\DatabaseAwareInterface;
use Joomla\Database\DatabaseAwareTrait;
use Joomla\Database\DatabaseDriver;
use Joomla\Event\Event;
use Joomla\Event\SubscriberInterface;

defined('_JEXEC') or die;

/**
 * Automatically replies to tickets based on a set of predefined rules
 *
 * @since 3.2.0
 */
class Autoreply extends CMSPlugin implements SubscriberInterface
{
	const atsCommandName = 'autoreply';

	/** @inheritdoc */
	protected $autoloadLanguage = true;

	/**
	 * @inheritDoc
	 */
	public static function getSubscribedEvents(): array
	{
		return [
			'onATSPost'         => 'onATSPost',
			'onAtsCronTaskInfo' => 'onAtsCronTaskInfo',
			'onAtsCronTask'     => 'onAtsCronTask',
		];
	}

	/**
	 * This handles the onATSPost event which is triggered every time a post is created or edited.
	 *
	 * We use this event to detect the first post of a new ticket and send auto-replies if we are configured to do so.
	 *
	 * @param   array  $info  An indexed array with post info. The post key contains an AtsTablePost object.
	 *
	 * @return  void
	 */
	public function onATSPost(Event $event)
	{
		/**
		 * @var bool        $new    Is this a new post?
		 * @var PostTable   $post   The post object saved to the database
		 * @var TicketTable $ticket The ticket object the post belongs to
		 */
		$new    = $event->getArgument('new');
		$post   = $event->getArgument('post');
		$ticket = $event->getArgument('ticket');

		// Is this feature enabled?
		if ($this->params->get('event', 0) != 1)
		{
			return;
		}

		// We only need to process NEW posts
		if (!$new)
		{
			return;
		}


		// Is this a new ticket?
		/** @var PostTable $firstPost */
		$firstPost   = $this->getFirstPost($ticket);
		$isNewTicket = $firstPost->getId() == $post->getId();

		// Only send auto-replies for NEW tickets
		if (!$isNewTicket)
		{
			return;
		}

		AutoReplyHelper::runAutoreplies([$ticket->getId()]);
	}

	/**
	 * Handles the ATS CRON tasks.
	 *
	 * @param   Event  $event
	 *
	 * @return  void
	 * @since   3.2.0
	 */
	public function onAtsCronTask(Event $event)
	{
		/**
		 * @var string $command The command name ATS CRON was asked to execute.
		 * @var array  $options Any options passed to this CRON job
		 */
		[$command, $options] = $event->getArguments();
		$result = $event->getArgument('result', []);

		if ($this->params->get('cron', 1) != 1)
		{
			return;
		}

		// Make sure we are calling the correct command
		if ($command != self::atsCommandName)
		{
			return;
		}

		$event->setArgument('result', array_merge($result, [true]));

		$timeLimit = array_key_exists('time_limit', $options) ? $options['time_limit'] : 86400;
		$timer     = new Timer($timeLimit);

		// Loop while we have tickets and we have not hit the time limit
		$result = AutoReplyHelper::runAutoreplies(null, $timer);

		Log::add(sprintf('Bots have replied to %d ticket(s)', $result), Log::DEBUG, 'ats.cron');
	}

	/**
	 * Returns information about the CRON task provided by this plugin
	 *
	 * @param   Event  $event  The event to handle
	 *
	 * @return  void
	 *
	 * @since   3.2.0
	 */
	public function onAtsCronTaskInfo(Event $event)
	{
		$result = $event->getArgument('result', []);
		$event->setArgument('result', array_merge($result, [
			[
				'command'     => self::atsCommandName,
				'label'       => 'PLG_ATS_' . $this->_name . '_TASK_LABEL',
				'description' => 'PLG_ATS_' . $this->_name . '_TASK_DESC',
			],
		]));
	}

	/**
	 * Returns the very first post of a ticket
	 *
	 * @param   TicketTable  $ticket
	 *
	 * @return  PostTable|null  Very first post; null if it does not exist.
	 *
	 * @since   5.0.0
	 */
	private function getFirstPost(TicketTable $ticket): ?PostTable
	{
		$allPosts = $ticket->posts($ticket->id);

		if (empty($allPosts))
		{
			return null;
		}

		$firstPost = array_shift($allPosts);

		return (is_object($firstPost) && ($firstPost instanceof PostTable)) ? $firstPost : null;
	}

}