<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2023 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Joomla\Plugin\Ats\Deletenotes\Extension;

use Akeeba\Component\ATS\Administrator\Helper\Timer;
use Akeeba\Component\ATS\Administrator\Table\TicketTable;
use Exception;
use Joomla\CMS\Application\CMSApplication;
use Joomla\CMS\Log\Log;
use Joomla\CMS\Plugin\CMSPlugin;
use Joomla\Database\DatabaseAwareInterface;
use Joomla\Database\DatabaseAwareTrait;
use Joomla\Database\DatabaseDriver;
use Joomla\Database\ParameterType;
use Joomla\Event\Event;
use Joomla\Event\SubscriberInterface;

defined('_JEXEC') or die;

/**
 * Deletes the Manager Notes of tickets when they are closed or unpublished.
 *
 * @since 3.2.0
 */
class Deletenotes extends CMSPlugin implements SubscriberInterface, DatabaseAwareInterface
{
	use DatabaseAwareTrait;

	const atsCommandName = 'deletenotes';

	/** @inheritdoc */
	protected $autoloadLanguage = true;

	public static function getSubscribedEvents(): array
	{
		return [
			'onAtsCronTask'                 => 'onAtsCronTask',
			'onAtsCronTaskInfo'             => 'onAtsCronTaskInfo',
			'onComAtsTableTicketAfterStore' => 'onComAtsTableTicketAfterStore',
		];
	}

	/**
	 * Handles the ATS CRON tasks.
	 *
	 * @param   Event  $event  The event to handle
	 *
	 * @return  void
	 * @since   3.2.0
	 */
	public function onAtsCronTask(Event $event): void
	{
		/**
		 * @var string $command The command to run
		 * @var array  $options Options to pass to the command
		 *
		 * Returns True on success, null if it's not the expected command name.
		 */
		[$command, $options] = $event->getArguments();
		$result = $event->getArgument('result', []);

		if ($this->params->get('cron', 1) != 1)
		{
			return;
		}

		// Make sure we are calling the correct command
		if ($command != self::atsCommandName)
		{
			return;
		}

		$event->setArgument('result', array_merge($result, [true]));

		$timeLimit = array_key_exists('time_limit', $options) ? $options['time_limit'] : 86400;
		$timer     = new Timer($timeLimit);
		/** @var DatabaseDriver $db */
		$db        = $this->getDatabase();

		// Loop while we have tickets and we have not hit the time limit
		while ($timer->getTimeLeft())
		{
			// Get 100 closed / unpublished tickets with manager notes
			Log::add('Getting up to 100 closed / unpublished tickets with manager notes', Log::DEBUG, 'ats.cron');

			$subQuery = $db->getQuery(true)
				->select($db->quoteName('id'))
				->from($db->quoteName('#__ats_tickets'))
				->where([
					$db->quoteName('enabled') . ' = 0',
					$db->quoteName('status') . ' = ' . $db->q('C'),
				], 'OR');


			$query = $db->getQuery(true)
				->select($db->quoteName('ticket_id'))
				->from($db->quoteName('#__ats_managernotes'))
				->where($db->quoteName('ticket_id') . ' IN (' . $subQuery . ')')
				->group($db->quoteName('ticket_id'))
				->order($db->quoteName('ticket_id') . ' ASC')
				->setLimit(0, 100);

			try
			{
				$ticketIDs = $db->setQuery($query)->loadColumn() ?: [];
			}
			catch (Exception $e)
			{
				Log::add(sprintf('Cannot perform database query: %s', $e->getMessage()), Log::ERROR, 'ats.cron');

				return;
			}

			$count = count($ticketIDs);

			Log::add(sprintf('Found %u ticket(s)', $count), Log::DEBUG, 'ats.cron');

			// No more tickets? Quit.
			if (!$count)
			{
				Log::add('No more tickets to process. Returning successfully.', Log::DEBUG, 'ats.cron');

				return;
			}

			// Delete the manager notes of these tickets
			Log::add(sprintf('Deleting manager notes from %u ticket(s)', $count), Log::DEBUG, 'ats.cron');

			$query = $db->getQuery(true)
				->delete($db->quoteName('#__ats_managernotes'))
				->whereIn($db->quoteName('ticket_id'), $ticketIDs, ParameterType::INTEGER);

			try
			{
				$db->setQuery($query)->execute();
			}
			catch (Exception $e)
			{
			}
		}

		Log::add(sprintf('I have reached the time limit of %u second(s). Stopping for now.', $timeLimit), Log::DEBUG, 'ats.cron');

		return;
	}

	/**
	 * Returns information about the CRON task provided by this plugin
	 *
	 * @param   Event  $event  The event to handle
	 *
	 * @return  void
	 *
	 * @since   3.2.0
	 */
	public function onAtsCronTaskInfo(Event $event): void
	{
		$result = $event->getArgument('result', []);
		$event->setArgument('result', array_merge($result, [
			[
				'command'     => self::atsCommandName,
				'label'       => 'PLG_ATS_' . $this->_name . '_TASK_LABEL',
				'description' => 'PLG_ATS_' . $this->_name . '_TASK_DESC',
			],
		]));
	}

	/**
	 * Delete the Manager Notes after a ticket is unpublished
	 *
	 * Hook into the onAfterStore event of ATS' TicketTable class. This event is fired automatically by ATS whenever we
	 * successfully store changes to a new or existing ticket.
	 *
	 * @param   Event  $event  The event to handle
	 *
	 * @return  void
	 * @since   3.2.0
	 */
	public function onComAtsTableTicketAfterStore(Event $event): void
	{
		/**
		 * @var TicketTable $ticket      The ticket table object which was saved
		 * @var bool        $updateNulls Whether the table object was asked to update null values
		 */
		[$ticket, $updateNulls] = $event->getArguments();

		if ($this->params->get('event', 0) != 1)
		{
			return;
		}

		// Only trigger when an existing ticket is unpublised (not enabled) or closed (status = C)
		$mustTrigger = $ticket->isNew() || !$ticket->enabled || ($ticket->status == 'C');

		if (!$mustTrigger)
		{
			return;
		}

		$notes = $ticket->managerNotes();

		if (empty($notes))
		{
			return;
		}

		foreach ($notes as $note)
		{
			try
			{
				$note->delete();
			}
			catch (Exception $e)
			{
				// No sweat, we'll do a traditional delete afterwards
			}
		}

		// Safe fallback
		$ticketId = $ticket->getId();
		/** @var DatabaseDriver $db */
		$db       = $this->getDatabase();
		$query    = $db->getQuery(true)
			->delete($db->quoteName('#__ats_managernotes'))
			->where($db->quoteName('ticket_id') . ' = :ticketId')
			->bind(':ticketId', $ticketId, ParameterType::INTEGER);

		try
		{
			$db->setQuery($query)->execute();
		}
		catch (Exception $e)
		{
			// Well, not much I can do if the DB has died on us.
		}
	}
}