<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2023 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Joomla\Plugin\Ats\Mailfetch\Extension;

defined('_JEXEC') or die;

use Akeeba\Component\ATS\Administrator\Helper\Timer;
use Exception;
use Joomla\CMS\Application\CMSApplication;
use Joomla\CMS\Application\ConsoleApplication;
use Joomla\CMS\Language\Text;
use Joomla\CMS\Log\Log;
use Joomla\CMS\Mail\Mail;
use Joomla\CMS\Mail\MailTemplate;
use Joomla\CMS\MVC\Factory\MVCFactoryAwareTrait;
use Joomla\CMS\MVC\Factory\MVCFactoryInterface;
use Joomla\CMS\Plugin\CMSPlugin;
use Joomla\Plugin\Ats\Mailfetch\Exceptions\TimeoutException;
use Joomla\Plugin\Ats\Mailfetch\Fetcher\EmailCheck;

/**
 * Download and process email to create new tickets and ticket replies
 *
 * @since 3.2.0
 */
class Mailfetch extends CMSPlugin
{
	use MVCFactoryAwareTrait;

	const atsCommandName = 'mailfetch';

	/** @inheritdoc */
	protected $autoloadLanguage = true;

	/** @inheritdoc */
	public function __construct(&$subject, $config, MVCFactoryInterface $factory)
	{
		$this->setMVCFactory($factory);

		parent::__construct($subject, $config);

		$autoloaderPath = __DIR__ . '/../../vendor/autoload.php';

		if (@file_exists($autoloaderPath))
		{
			require_once $autoloaderPath;
		}
	}

	/**
	 * Handles the ATS CRON tasks.
	 *
	 * @param   string  $command  The command name ATS CRON was asked to execute.
	 * @param   array   $options  Any options passed to this CRON job
	 *
	 * @return  bool|null  True on success, null if it's not the expected command name.
	 * @since   3.2.0
	 */
	public function onAtsCronTask($command, array $options = [])
	{
		// Make sure we are calling the correct command
		if ($command != self::atsCommandName)
		{
			return null;
		}

		$timeLimit = array_key_exists('time_limit', $options) ? $options['time_limit'] : 86400;
		$timer     = new Timer($timeLimit);

		$emailCheck = new EmailCheck($this->getMVCFactory());

		// Load Joomla global language file. Needed to print nice warning messages (ie attachment upload error)
		$lang = $this->getApplication()->getLanguage();
		$lang->load('lib_joomla', JPATH_ADMINISTRATOR);

		try
		{
			$emailCheck->checkEmail($timer);

			Log::add('Done retrieving email messages', Log::DEBUG, 'ats.cron');
		}
		catch (TimeoutException $e)
		{
			Log::add(sprintf('I have reached the time limit of %u second(s). Stopping for now.', $timeLimit), Log::DEBUG, 'ats.cron');
		}
		catch (Exception $e)
		{
			Log::add('Failed to retrieve email messages', Log::DEBUG, 'ats.cron');
		}

		return true;
	}

	/**
	 * Returns information about the CRON task provided by this plugin
	 *
	 * @return  array
	 *
	 * @since   3.2.0
	 */
	public function onAtsCronTaskInfo()
	{
		return [
			'command'     => self::atsCommandName,
			'label'       => 'PLG_ATS_' . $this->_name . '_TASK_LABEL',
			'description' => 'PLG_ATS_' . $this->_name . '_TASK_DESC',
		];
	}

	public function onMailBeforeRendering($templateId, MailTemplate $template)
	{
		// Make sure the Reply by Email feature is enabled
		if ($this->params->get('replybyemail', 0) != 1)
		{
			return;
		}

		if (empty($template) || strpos($templateId, 'com_ats.') !== 0)
		{
			return;
		}

		// If the "Only for managers" option is enabled make sure it's a mail sent to a category manager
		$emailReplyOnlyForManager = $this->params->get('emailadminonly', 0);

		if ($emailReplyOnlyForManager && strpos($templateId, 'com_ats.manager_') !== 0)
		{
			return;
		}

		// Get the template data. Joomla makes this unnecessarily hard :(
		$refObject = new \ReflectionObject($template);
		$refProp   = $refObject->getProperty('data');
		$refProp->setAccessible(true);
		$data     = $refProp->getValue($template);
		$ticketId = $data['id'] ?? 0;

		if (empty($ticketId))
		{
			return;
		}

		// Get the reply line's content
		$replyLine = sprintf(
			"!-!- %s {ticketid:%s} -!-!",
			Text::_('COM_ATS_POST_LBL_REPLYABOVE'),
			$ticketId
		);

		$template->addTemplateData([
			'replyafter'      => $replyLine,
			'replyafter_html' => "</p>$replyLine<p>",
		]);

		// Add a custom header, too. Again, Joomla is incapable of anything even remotely close to useful.
		$refProp = $refObject->getProperty('mailer');
		$refProp->setAccessible(true);
		/** @var Mail $mailer */
		$mailer = $refProp->getValue($template);

		try
		{
			$mailer->addCustomHeader('X-ATS-ticketid:' . $ticketId);
		}
		catch (Exception $e)
		{
			// Joomla's Mail died on us. No problem, we still have the reply line.
		}

		$refProp->setValue($template, $mailer);
	}
}