<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2023 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Joomla\Plugin\Ats\Removeattachments\Extension;

defined('_JEXEC') or die;

use Akeeba\Component\ATS\Administrator\Helper\Timer;
use Akeeba\Component\ATS\Administrator\Table\AttachmentTable;
use Exception;
use Joomla\CMS\Factory;
use Joomla\CMS\Log\Log;
use Joomla\CMS\Plugin\CMSPlugin;
use Joomla\Database\DatabaseAwareInterface;
use Joomla\Database\DatabaseAwareTrait;
use Joomla\Database\DatabaseDriver;
use Joomla\Database\ParameterType;
use Joomla\Event\Event;
use Joomla\Event\SubscriberInterface;

/**
 * Deletes the Manager Notes of tickets when they are closed or unpublished.
 *
 * @since 3.2.0
 */
class Removeattachments extends CMSPlugin implements SubscriberInterface, DatabaseAwareInterface
{
	use DatabaseAwareTrait;

	const atsCommandName = 'removeattachments';

	/** @inheritdoc */
	protected $autoloadLanguage = true;

	public static function getSubscribedEvents(): array
	{
		return [
			'onAtsCronTask'     => 'onAtsCronTask',
			'onAtsCronTaskInfo' => 'onAtsCronTaskInfo',
		];
	}

	/**
	 * Handles the ATS CRON tasks.
	 *
	 * @param   Event  $event
	 *
	 * @throws  Exception
	 * @since   3.2.0
	 */
	public function onAtsCronTask(Event $event): void
	{
		/**
		 * @var string $command The command to run
		 * @var array  $options Options to pass to the command
		 *
		 * Returns True on success, null if it's not the expected command name.
		 */
		[$command, $options] = $event->getArguments();
		$result = $event->getArgument('result', []);

		// Make sure we are calling the correct command
		if ($command != self::atsCommandName)
		{
			return;
		}

		$event->setArgument('result', array_merge($result, [true]));

		$timeLimit    = array_key_exists('time_limit', $options) ? $options['time_limit'] : 86400;
		$timer        = new Timer($timeLimit);
		$onlyClosed   = $this->params->get('only_closed', 1) == 1;
		$totalDeleted = 0;

		$max_age = $this->params->get('maxattachmentage', 30);
		$until   = time() - $max_age * 3600 * 24;
		$jUntil  = clone Factory::getDate($until);

		// Loop while we have tickets, and we have not hit the time limit
		while ($timer->getTimeLeft())
		{
			Log::add(sprintf("Getting up to 100 attachments%s", $onlyClosed ? ' of closed / disabled tickets' : ''), Log::DEBUG, 'ats.cron');

			/** @var DatabaseDriver $db */
			$db       = $this->getDatabase();
			$untilSql = $jUntil->toSql();
			$query    = $db->getQuery(true)
				->select($db->quoteName('a.id'))
				->from($db->quoteName('#__ats_attachments', 'a'))
				->where($db->quoteName('a.created') . ' <= :untilSql')
				->where($db->quoteName('a.enabled') . ' = 1')
				->setLimit(0, 100)
				->bind(':untilSql', $untilSql, ParameterType::STRING);

			// Filter for enabled tickets when necessary
			if ($onlyClosed)
			{
				$query
					->innerJoin($db->quoteName('#__ats_posts', 'p'),
						$db->quoteName('p.id') . ' = ' . $db->quoteName('a.post_id')
					)
					->innerJoin($db->quoteName('#__ats_tickets', 't'),
						$db->quoteName('t.id') . ' = ' . $db->quoteName('p.ticket_id')
					)
					->extendWhere('AND', [
						$db->quoteName('status') . ' = ' . $db->quote('C'),
						$db->quoteName('t.enabled') . ' = 0',
					], 'OR');
			}

			$attachmentIds = $db->setQuery($query)->loadAssocList() ?: [];

			if (count($attachmentIds) == 0)
			{
				if ($totalDeleted)
				{
					Log::add(sprintf('Deleted a total of %d attachments', $totalDeleted), Log::INFO, 'ats.cron');
				}
				else
				{
					Log::add('No attachments to delete were found', Log::INFO, 'ats.cron');
				}

				return;
			}

			Log::add(sprintf('Found %d attachments to process', count($attachmentIds)), Log::DEBUG, 'ats.cron');

			$attachmentTable = new AttachmentTable($db);

			foreach ($attachmentIds as $id)
			{
				if ($timer->getTimeLeft() <= 0)
				{
					Log::add(sprintf('Deleted %d attachments in this batch', $deleted), Log::DEBUG, 'ats.cron');

					break 2;
				}

				Log::add(sprintf(
					'Deleting attachment #%d', $id
				), Log::INFO, 'ats.cron');

				if ($attachmentTable->delete($id))
				{
					$deleted++;
					$totalDeleted++;
				}
			}

			Log::add(sprintf('Deleted %d attachments in this batch', $deleted), Log::DEBUG, 'ats.cron');
		}

		Log::add(sprintf('I have reached the time limit of %u second(s). Stopping for now.', $timeLimit), Log::DEBUG, 'ats.cron');
		Log::add(sprintf('Deleted a total of %d attachments. More attachments to delete exist (I ran out of time).', $totalDeleted), Log::INFO, 'ats.cron');
	}

	/**
	 * Returns information about the CRON task provided by this plugin
	 *
	 * @return  void
	 *
	 * @since   3.2.0
	 */
	public function onAtsCronTaskInfo(Event $event)
	{
		$result = $event->getArgument('result', []);
		$event->setArgument('result', array_merge($result, [
			[
				'command'     => self::atsCommandName,
				'label'       => 'PLG_ATS_' . $this->_name . '_TASK_LABEL',
				'description' => 'PLG_ATS_' . $this->_name . '_TASK_DESC',
			],
		]));
	}
}