<?php
/**
 * @package   ats
 * @copyright Copyright (c)2011-2023 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Joomla\Plugin\Ats\Mailfetch\Fetcher;

defined('_JEXEC') or die;

use Akeeba\Component\ATS\Administrator\Helper\AddonEmails;
use Akeeba\Component\ATS\Administrator\Helper\Attachment;
use Akeeba\Component\ATS\Administrator\Helper\Filter;
use Akeeba\Component\ATS\Administrator\Helper\Logger;
use Akeeba\Component\ATS\Administrator\Helper\Permissions;
use Akeeba\Component\ATS\Administrator\Helper\TemplateEmails;
use Akeeba\Component\ATS\Administrator\Helper\Text2Html;
use Akeeba\Component\ATS\Administrator\Helper\Timer;
use Akeeba\Component\ATS\Administrator\Mixin\ComponentObjectTrait;
use Akeeba\Component\ATS\Administrator\Mixin\ControllerNewPostTrait;
use Akeeba\Component\ATS\Administrator\Model\TicketModel;
use Akeeba\Component\ATS\Administrator\Table\AttachmentTable;
use Akeeba\Component\ATS\Administrator\Table\PostTable;
use Akeeba\Component\ATS\Administrator\Table\TicketTable;
use Exception;
use Joomla\CMS\Application\ApplicationHelper;
use Joomla\CMS\Categories\Categories;
use Joomla\CMS\Component\ComponentHelper;
use Joomla\CMS\Factory;
use Joomla\CMS\Filesystem\File;
use Joomla\CMS\Log\Log as JLog;
use Joomla\CMS\MVC\Factory\MVCFactoryAwareTrait;
use Joomla\CMS\MVC\Factory\MVCFactoryInterface;
use Joomla\CMS\Plugin\PluginHelper;
use Joomla\CMS\Uri\Uri;
use Joomla\CMS\User\User;
use Joomla\Database\DatabaseAwareTrait;
use Joomla\Plugin\Ats\Mailfetch\Exceptions\TimeoutException;
use Joomla\Registry\Registry;
use Joomla\Utilities\ArrayHelper;
use RuntimeException;

class EmailCheck
{
	use MVCFactoryAwareTrait;
	use DatabaseAwareTrait;
	use ControllerNewPostTrait;
	use ComponentObjectTrait;

	/**
	 * Have I already registered a logger for the email checks?
	 *
	 * @var   bool
	 * @since 3.2.0
	 */
	protected static $hasRegisteredLogger = false;

	/**
	 * The mail fetcher backend
	 *
	 * @var   MailClientInterface
	 * @since 3.2.0
	 */
	protected $backend;

	/**
	 * The options of the mailfetch plugin
	 *
	 * @var   Registry
	 * @since 3.2.0
	 */
	protected $pluginOptions;

	/**
	 * @inheritDoc
	 */
	public function __construct(MVCFactoryInterface $factory)
	{
		$this->setMVCFactory($factory);
		$this->setDatabase(Factory::getContainer()->get('DatabaseDriver'));

		// Register a log file for email fetching
		if (!self::$hasRegisteredLogger)
		{
			Logger::registerFileLogger('ats.emails', 'ats_email_check.php');

			self::$hasRegisteredLogger = true;
		}

		// Populate the plugin options
		$this->populatePluginOptions();
	}

	/**
	 * Check for new email and process the messages
	 *
	 * @param   Timer  $timer  The timer object which controls execution time
	 *
	 * @since   3.2.0
	 */
	public function checkEmail(Timer $timer)
	{
		JLog::add("Beginning checking for emails", JLog::INFO, 'ats.emails');

		// Get parameters
		$replyByEmail = $this->getPluginOption('replybyemail', '0');
		$newByEmail   = $this->getPluginOption('newbyemail', '0');

		// Bail out if no email feature is set at all
		if (!$replyByEmail && !$newByEmail)
		{
			JLog::add("Reply by Email and Create Ticket by Email are both disabled.", JLog::WARNING, 'ats.emails');

			return;
		}

		// Get the mail backend
		try
		{
			JLog::add("Attempting to connect to mail server", JLog::INFO, 'ats.emails');

			$mailbox = $this->getMailbox();
		}
		catch (Exception $e)
		{
			JLog::add("Mail server connection failed.", JLog::ERROR, 'ats.emails');
			JLog::add($e->getMessage(), JLog::DEBUG, 'ats.emails');

			return;
		}

		// Check for emails since the last mail check
		$lastCheckDate = $this->getLastTimestamp();

		// If there is no last check date assume it was 24 hours ago
		if (empty($lastCheckDate))
		{
			$lastCheckDate = time() - 86400;
		}

		// Set the last check date to prevent this loop running in parallel several times
		$this->setLastTimestamp();

		$formattedSinceDate = date("r", $lastCheckDate - 24 * 3600);
		JLog::add(sprintf("Checking for unread email sent after %s", $formattedSinceDate), JLog::INFO, 'ats.emails');

		try
		{
			$mailbox->iterateUnread(function (MailMessage $email) use ($timer) {
				static $iterations = 0;

				// Check if we have enough time left to process another email. DO NOT REMOVE.
				if (($iterations > 0) && ($timer->getTimeLeft() <= 0.01))
				{
					throw new TimeoutException();
				}

				$debugLine = sprintf("%s. %s <%s> — %s", $email->mID, $email->fromName, $email->fromEmail, $email->subject);
				JLog::add($debugLine, JLog::INFO, 'ats.emails');

				$iterations++;

				return $this->processEmail($email);
			}, $formattedSinceDate);
		}
		catch (TimeoutException $e)
		{
			// Reset last check time so the next page load / CRON task will pick up where we left off.
			$this->setLastTimestamp($lastCheckDate);

			// The rethrow the exception so the plugin code knows we ran out of time
			throw $e;
		}
		catch (Exception $e)
		{
			JLog::add(sprintf("Failed to process email messages: [#%d] %s", $e->getCode(), $e->getMessage()), JLog::ERROR, 'ats.emails');
			JLog::add(sprintf("%s::L%u", $e->getFile(), $e->getLine()), JLog::ERROR, 'ats.emails');

			foreach (explode("\n", $e->getTraceAsString()) as $line)
			{
				JLog::add($line, JLog::ERROR, 'ats.emails');
			}
		}
	}

	/**
	 * Get the timestamp of the last email check
	 *
	 * @return int
	 */
	public function getLastTimestamp()
	{
		return $this->getPluginOption('email_lastcheck', 0);
	}

	/**
	 * Get a configuration parameter of the plg_ats_mailfetch plugin.
	 *
	 * @param   string  $key
	 * @param   mixed   $default
	 *
	 * @return  mixed
	 * @since   3.2.0
	 */
	public function getPluginOption($key, $default = null)
	{
		return $this->pluginOptions->get($key, $default);
	}

	/**
	 * Set the timestamp of the last email check to now
	 *
	 * @return  void
	 */
	public function setLastTimestamp($timestamp = null)
	{
		if (empty($timestamp))
		{
			$timestamp = time();
		}

		$this->pluginOptions->set('email_lastcheck', $timestamp);

		$this->getExtensionObject()
			->getComponentParametersService()
			->save($this->pluginOptions, 'plg_ats_mailfetch');
	}

	/**
	 * Populate $this->pluginOptions with the options of the plg_ats_mailfetch plugin
	 *
	 * @return  void
	 * @since   3.2.0
	 */
	protected function populatePluginOptions()
	{
		$plugin = PluginHelper::getPlugin('ats', 'mailfetch');

		$this->pluginOptions = new Registry($plugin->params);
	}

	/**
	 * Extract the attachments from the email message and store them in ATS. Returns an array with the generated ATS
	 * attachment IDs.
	 *
	 * @param   MailMessage  $email   The email message parts
	 * @param   int          $userID  The user ID applicable when saving the extracted attachments
	 *
	 * @return  array  The IDs of the attachments
	 *
	 * @throws  Exception
	 * @since   3.2.0
	 */
	private function extractAttachments(MailMessage $email, int $userID): array
	{
		if (!count($email->attachments))
		{
			return [];
		}

		// Get the temporary directory
		$app     = Factory::getApplication();
		$tempdir = $app->get('tmp_path');
		$ret     = [];

		foreach ($email->attachments as $attachment)
		{
			$size    = strlen($attachment['data']);
			$tmpName = tempnam($tempdir, 'atsup');
			$res     = @file_put_contents($tmpName, $attachment['data']);

			if ($res === false)
			{
				continue;
			}

			$fakeFileStructure = [
				// Joomla rejects files with spaces in their name. So I'll proactively substitute them with underscores.
				'name'     => File::makeSafe($attachment['filename']),
				'type'     => $attachment['mime'],
				'tmp_name' => $tmpName,
				'error'    => '',
				'size'     => $size,
			];

			try
			{
				$fileDefinition = Attachment::uploadFile($fakeFileStructure, false);
			}
			catch (Exception $e)
			{
				// Don't die.
				$fileDefinition = false;
			}

			if ($fileDefinition === false)
			{
				@unlink($tmpName);

				continue;
			}

			$fileDefinition['created_by'] = $userID;
			$fileDefinition['created']    = (clone Factory::getDate())->toSql();
			$fileDefinition['enabled']    = 1;
			$fileDefinition['post_id']    = 0;

			/** @var AttachmentTable $attachmentTable */
			$attachmentTable = $this->getMVCFactory()->createTable('Attachment', 'Administrator');
			$attachmentTable->reset();

			// Save the attachment; if saving fails, remove the uploaded file
			if (!$attachmentTable->save($fileDefinition))
			{
				$toDelete = $attachmentTable->getAbsoluteFilename();

				if (!@unlink($toDelete))
				{
					File::delete($toDelete);
				}
			}

			@unlink($tmpName);

			$ret[] = $attachmentTable->getId();
		}

		return $ret;
	}

	/**
	 * Extracts the part of the message which is above the reply line
	 *
	 * @param   MailMessage  $email
	 *
	 * @return  string
	 * @since   3.2.0
	 */
	private function extractMessage(MailMessage $email): string
	{
		// If it's a plain text email I need to convert it to HTML first.
		$body = $email->body;

		if (!$email->isHtml)
		{
			$body = Text2Html::convert($body);
		}

		// If there's a body tag return the content in the body
		$hasBody = preg_match_all('#(<body[^>]*>(.*?)</body>)#ims', $body, $matches);

		if ($hasBody)
		{
			$body = implode('', $matches[2]);
		}

		$parts = explode('!-!-', $body);

		return $parts[0];
	}

	/**
	 * Gets the category ID based on the email address of the user
	 *
	 * @staticvar null|array $catmap Holds a map of emails to categories
	 *
	 * @param   string|null  $email  Email address to check
	 *
	 * @return  int  The category ID, or 0 if none is found
	 * @since     3.2.0
	 */
	private function getCategoryIDFromEmail(?string $email): int
	{
		static $catMap = null;

		if (empty($email))
		{
			return 0;
		}

		if (is_null($catMap))
		{
			$catMap     = [];
			$db         = $this->getDatabase();
			$query      = $db->getQuery(true)
				->select($db->quoteName('id'))
				->from($db->quoteName('#__categories'))
				->where($db->quoteName('extension') . ' = ' . $db->quote('com_ats'));
			$ids        = array_unique(ArrayHelper::toInteger($db->setQuery($query)->loadColumn() ?: []));
			$catService = Categories::getInstance('com_ats');

			foreach ($ids as $id)
			{
				$category = $catService->get($id);

				if (empty($category) || ($category->id != $id))
				{
					continue;
				}

				$params   = $category->getParams();
				$catEmail = $params->get('category_email', '');

				if (empty($catEmail))
				{
					continue;
				}

				$catEmail = strtolower($catEmail);

				if (array_key_exists($catEmail, $catMap))
				{
					continue;
				}

				$catMap[$catEmail] = $category->id;
			}
		}

		$email = strtolower($email);

		if (array_key_exists($email, $catMap))
		{
			return $catMap[$email];
		}

		return 0;
	}

	/**
	 * Return the mail client backend used to fetch and process incoming mail messages
	 *
	 * @return  MailClientInterface
	 *
	 * @throws  Exception  When connecting to the mail server fails for any reason
	 * @since   3.2.0
	 */
	private function getMailbox(): MailClientInterface
	{
		if (is_null($this->backend))
		{
			$mailFetcherParams = [
				'mailbox_type'  => $this->getPluginOption('mailboxtype', 'imap'),
				'server'        => $this->getPluginOption('email_server', ''),
				'port'          => $this->getPluginOption('email_port', ''),
				'tls'           => $this->getPluginOption('email_tls', ''),
				'username'      => $this->getPluginOption('email_username', ''),
				'password'      => $this->getPluginOption('email_password', ''),
				'folder'        => $this->getPluginOption('email_folder', 'INBOX'),
				'access_token'  => '',
				'refresh_token' => '',
				'client_id'     => '',
				'secret_key'    => '',
				'deleteAfter'   => $this->getPluginOption('email_delete', '1') == 1,
				'expunge'       => true,
			];

			switch ($mailFetcherParams['mailbox_type'])
			{
				case 'gmail':
					$mailFetcherParams = array_merge($mailFetcherParams, [
						'access_token'  => $this->getPluginOption('access_token', ''),
						'refresh_token' => $this->getPluginOption('refresh_token', ''),
						'client_id'     => $this->getPluginOption('gmail_client_id', ''),
						'secret_key'    => $this->getPluginOption('gmail_api_secret', ''),
					]);
					break;

				case 'microsoft':
					$mailFetcherParams = array_merge($mailFetcherParams, [
						'access_token'  => $this->getPluginOption('access_token', ''),
						'refresh_token' => $this->getPluginOption('refresh_token', ''),
						'tenant_id'     => $this->getPluginOption('microsoft_tenant', ''),
						'client_id'     => $this->getPluginOption('microsoft_client_id', ''),
						'secret_key'    => $this->getPluginOption('microsoft_secret_value', ''),
					]);
					break;
			}

			$this->backend = new Horde($mailFetcherParams);
		}

		return $this->backend;
	}

	/**
	 * Returns the ticket ID found inside an email
	 *
	 * @param   MailMessage  $email
	 *
	 * @return  null|int  Null if no ID is found, else the ticket ID
	 */
	private function getTicketID(MailMessage $email): ?int
	{
		// Ticket ID from the header (if any)
		$ticketId = empty($email->ticketID) ? 0 : $email->ticketID;

		// Ticket ID from the body (stated in "reply below this line")
		$hasTicketId  = preg_match('/{ticketid:(\d+)}/ism', $email->body, $matches);
		$bodyTicketId = $hasTicketId ? $matches[1] : 0;

		// Get the ticket ID that's non zero. If there are two, keep the correct one.
		$ticketId = max($ticketId, $bodyTicketId);

		return empty($ticketId) ? null : (int) $ticketId;
	}

	/**
	 * Get the number of tickets already created with an email that has the same UID as the one in $email
	 *
	 * @param   MailMessage  $email  The email message to get the UID from
	 *
	 * @return  int
	 *
	 * @since   3.2.0
	 */
	private function getUIDCount(MailMessage $email): int
	{
		$db    = $this->getDatabase();
		$query = $db->getQuery(true)
			->select('COUNT(*)')
			->from($db->qn('#__ats_posts'))
			->where($db->qn('email_uid') . ' = ' . $db->q($email->uID));

		return (int) ($db->setQuery($query)->loadResult() ?: 0);
	}

	/**
	 * Does the message have a "reply above this line" line?
	 *
	 * @param   string  $body
	 *
	 * @return  bool
	 * @since   3.2.0
	 */
	private function hasReplyAboveLine(string $body): bool
	{
		$regex  = '#!-!-(.*)-!-!#s';
		$result = preg_match_all($regex, $body, $matches);

		return ($result !== false) && ($result > 0);
	}

	/**
	 * Perform basic checks, i.e. whether this email can be considered at all.
	 *
	 * @param   MailMessage  $email       The email message to process
	 * @param   int          $categoryID  The ATS category ID
	 * @param   TicketTable  $ticket      The ticket this message will be processed into
	 * @param   string       $from        The email address of the sender
	 * @param   User         $user        The user record of the sender
	 *
	 * @return  bool
	 * @throws  Exception
	 * @since   3.2.0
	 */
	private function passedBasicChecks(MailMessage $email, int $categoryID, TicketTable $ticket, string &$from, ?User &$user): bool
	{
		// Get a preliminary user from the email address. If no user is found we will get a blank User (Guest) user.
		$from = $email->fromEmail;
		$uid  = AddonEmails::findUserIdByAddonEmail($from);
		$user = $uid ? Permissions::getUser($uid) : new User();

		$catService = Categories::getInstance('com_ats');
		$category   = null;

		if (!empty($categoryID))
		{
			$category = $catService->get($categoryID);
		}

		// Do I have a valid category?
		if (empty($category) || ($category->id != $categoryID))
		{
			JLog::add("Cannot load the ticket category. This is a configuration error.", JLog::ERROR, 'ats.emails');
			JLog::add("Sending the mailgateway-noaccess email to the user.", JLog::DEBUG, 'ats.emails');

			// Send an email telling the user he has no access
			$this->sendErrorEmail('mailgateway_noaccess', $from, [
				'name'     => $user->name,
				'username' => $user->username,
				'user_id'  => $user->id,
			]);

			return false;
		}

		/**
		 * Does the user has access to the category?
		 *
		 * This requires two checks.
		 *
		 * First we try to load the category record passing the user ID. This applies the View Access Levels. If we get
		 * null instead of a record the View Access Levels forbid the user to even view the category, let alone post.
		 *
		 * Then we get the core.create privilege of the user for the category. This tells us if he can post to the
		 * category.
		 */
		$canCreate = false;
		$category  = null;

		if (!empty($categoryID))
		{
			JLog::add(sprintf("Attempting to load ticket category #%d", $categoryID), JLog::DEBUG, 'ats.emails');
			$category = $catService->get($categoryID);
		}

		// Get ACL permissions and make sure we can post here
		if (!is_null($category))
		{
			$perms     = Permissions::getAclPrivileges($category->id, $user->id);
			$canCreate = $perms['core.create'];
		}

		/**
		 * Can we actually create a ticket? There are three checks to perform IN THIS PRECISE ORDER:
		 * 1- No User, but guest is allowed. This will give us a User if needed
		 * 2- If we got a User at all. There's no point in continuing without
		 * 3- Check if we have enough permissions on the category
		 */

		// If the user is not found AND the category allows Guest to post, let's try creating a new user
		if (!$user->id && $canCreate)
		{
			JLog::add('This is a Guest User ticket (new ticket from a user we have not seen before).', JLog::DEBUG, 'ats.emails');

			JLog::add("Trying to create a new user from this email address.", JLog::DEBUG, 'ats.emails');

			$fullName = $email->fromName ?: $from;

			if ($fullName != $from)
			{
				$fullName = str_replace('@', '.', $fullName);
				$username = ApplicationHelper::stringURLSafe($fullName);
			}
			else
			{
				$username = $from;
			}

			/** @var TicketModel $ticketModel */
			$ticketModel = $this->getMVCFactory()->createModel('Ticket', 'Administrator', ['ignore_request' => true]);
			$userId      = $ticketModel->createUser($fullName, $username, $from);

			if ($userId === false)
			{
				JLog::add(sprintf('Guest User ticket failed. Cannot create a new user from email address %s.', $from), JLog::ERROR, 'ats.emails');

				$userId = -1;
			}

			if ($userId > 0)
			{
				$user = Permissions::getUser($userId);
				$from = $user->email;
			}
		}

		// Catch the case of not having a valid user record yet.
		if (!$user->id)
		{
			JLog::add("No user record can be found from the sender's email address.", JLog::ERROR, 'ats.emails');
			JLog::add("Sending the mailgateway-invaliduser email and aborting.", JLog::DEBUG, 'ats.emails');

			// Maybe send an email telling the user he screwed up if there's a valid ticket ID?
			$this->sendErrorEmail('mailgateway_invaliduser', $from);

			return false;
		}

		// No category found (wrong or no access to it) or the user can't create a ticket
		if (!$canCreate)
		{
			if (is_null($category))
			{
				JLog::add(sprintf("User %s is not allowed to view the category because of View Access Level settings.", $user->id), JLog::ERROR, 'ats.emails');
			}
			else
			{
				JLog::add(sprintf("User %s can view the category but is not allowed to create new tickets in it.", $user->id), JLog::ERROR, 'ats.emails');
			}

			JLog::add("Sending the mailgateway-newreplydisabled email to the user", JLog::DEBUG, 'ats.emails');

			// Send an email telling the user that replies by email are not allowed
			$this->sendErrorEmail('mailgateway_newreplydisabled', $from, [
				'name'     => $user->name,
				'username' => $user->username,
				'user_id'  => $user->id,
			]);

			return false;
		}

		/**
		 * The user account may have been found using an add-on email. We have to normalize this to the Joomla! user
		 * account email. This is the only way we can email ALL of the addresses (canonical and add-on) of the user.
		 */
		$cParams        = ComponentHelper::getParams('com_ats');
		$from           = $user->email;
		$newByEmail     = $this->getPluginOption('newbyemail', 0);
		$replyByEmail   = $this->getPluginOption('replybyemail', '0');
		$emailAdminOnly = $this->getPluginOption('emailadminonly', 0);
		$noNewTickets   = $cParams->get('nonewtickets', 0);
		$noReplies      = $cParams->get('noreplies', 0);
		$isManager      = Permissions::isManager($category->id, $user->id);
		$isNewTicket    = $ticket->getId() <= 0;

		// Replies only allowed by a manager and this is not a manager
		if (!$isManager && $emailAdminOnly)
		{
			JLog::add("Reply by Email is only allowed for category managers (staff) but the person emailing us is not a manager.", JLog::ERROR, 'ats.emails');
			JLog::add("Sending the mailgateway-noaccess email and aborting.", JLog::DEBUG, 'ats.emails');

			// Send an email telling the user he has no access
			$this->sendErrorEmail('mailgateway_noaccess', $from, [
				'name'     => $user->name,
				'username' => $user->username,
				'user_id'  => $user->id,
			]);

			return false;
		}

		// Is New Ticket by Email globally disabled?
		if (!$ticket->getId() && !$newByEmail)
		{
			JLog::add("New Ticket by Email is not allowed and this is not a recognizable reply to an existing ticket.", JLog::ERROR, 'ats.emails');
			JLog::add("Sending the mailgateway-newticketdisabled email and aborting.", JLog::DEBUG, 'ats.emails');

			// New ticket creation not allowed in config
			$this->sendErrorEmail('mailgateway_newticketdisabled', $from, [
				'name'     => $user->name,
				'username' => $user->username,
				'user_id'  => $user->id,
			]);

			return false;
		}

		// Trying to create new ticket but we are not allowed to create new tickets globally and this is not a manager
		if ($isNewTicket && $noNewTickets && !$isManager)
		{
			JLog::add("'No new tickets' is set in the component configuration.", JLog::ERROR, 'ats.emails');
			JLog::add("Sending the mailgateway-nonewtickets email and aborting", JLog::DEBUG, 'ats.emails');

			// Send an email to the user telling him that the No New Tickets global switch is enabled
			$this->sendErrorEmail('mailgateway_nonewtickets', $from, [
				'name'     => $user->name,
				'username' => $user->username,
				'user_id'  => $user->id,
			]);

			return false;
		}

		// Existing ticket. Can we post the reply?
		if (!$isNewTicket && !$replyByEmail)
		{
			JLog::add("Reply By Email is disabled in the component's configuration.", JLog::ERROR, 'ats.emails');
			JLog::add("Sending the mailgateway-newreplydisabled email and aborting", JLog::DEBUG, 'ats.emails');

			// New ticket creation not allowed in config
			$this->sendErrorEmail('mailgateway_newreplydisabled', $from, [
				'name'     => $user->name,
				'username' => $user->username,
				'user_id'  => $user->id,
			]);

			return false;
		}

		// Replying to an existing ticket. Do I have the special ticket line?
		if (!$isNewTicket && !$this->hasReplyAboveLine($email->body))
		{
			JLog::add("Can't find the 'reply above the line' text", JLog::ERROR, 'ats.emails');
			JLog::add("Sending the mailgateway-noreplyline email and aborting", JLog::DEBUG, 'ats.emails');

			$this->sendErrorEmail('mailgateway_noreplyline', $from, [
				'name'     => $user->name,
				'username' => $user->username,
				'user_id'  => $user->id,
			]);

			return false;
		}

		// Existing ticket, we are not allowed to create new replies and we are not a manager
		if (!$isNewTicket && $noReplies && !$isManager)
		{
			JLog::add("User is not a manager and we are not allowed to create new replies in the component's configuration.", JLog::ERROR, 'ats.emails');
			JLog::add("Sending the mailgateway-nonewreplies email and aborting", JLog::DEBUG, 'ats.emails');

			// Send an email to the user telling him that the No New Replies global switch is enabled
			$this->sendErrorEmail('mailgateway_nonewreplies', $from, [
				'name'     => $user->name,
				'username' => $user->username,
				'user_id'  => $user->id,
			]);

			return false;
		}

		// Existing ticket, not a manager, not the user who created this ticket. Invalid reply.
		if (!$isNewTicket && !$isManager && ($user->id != $ticket->created_by))
		{
			JLog::add("The user emailing the reply to the ticket is not the user who created it.", JLog::ERROR, 'ats.emails');
			JLog::add("Sending the mailgateway-noaccess email and aborting", JLog::DEBUG, 'ats.emails');

			// Illegal post by a non-admin, non-owner user. Send an email telling the user he has no access.
			$this->sendErrorEmail('mailgateway_noaccess', $from, [
				'name'     => $user->name,
				'username' => $user->username,
				'user_id'  => $user->id,
			]);

			return false;
		}

		// All basic checks completed successfully
		return true;
	}

	/**
	 * Process an email message
	 *
	 * @param   MailMessage  $email  The email message to process
	 *
	 * @return  bool  True if it's successfully processed
	 *
	 * @throws  Exception
	 * @since   3.2.0
	 */
	private function processEmail(MailMessage $email): bool
	{
		// Make sure we are not processing the same email twice (only if requested to)
		$trackUID = $this->getPluginOption('track_email_id', 0) != 0;

		if ($trackUID && $this->getUIDCount($email) > 0)
		{
			JLog::add(sprintf("Skipping already processed email with UID #%s", $email->uID), JLog::NOTICE, 'ats.emails');

			return false;
		}

		// Get the relevant component configuration
		$defaultCategory = $this->getPluginOption('email_default_category', '0');

		// Get the ticket ID, report and validate it
		$ticketID = $this->getTicketID($email);

		if (is_null($ticketID))
		{
			JLog::add('No TicketID found: this email is not a reply to an existing ticket', JLog::DEBUG, 'ats.emails');
		}
		else
		{
			JLog::add(sprintf("TicketID found: this email is a reply to the existing ticket #%d", $ticketID), JLog::DEBUG, 'ats.emails');
		}

		/** @var TicketTable $ticket */
		$ticket = $this->getMVCFactory()->createTable('Ticket', 'Administrator');

		if (!is_null($ticketID))
		{
			if (!($ticket->load($ticketID)))
			{
				JLog::add(sprintf("Cannot find ticket #%d. Skipping this email.", $ticketID), JLog::ERROR, 'ats.emails');

				return false;
			}
		}

		// Get the category ID
		if (!$ticket->getId())
		{
			JLog::add("New ticket. Trying to determine category from the email address it was sent to.", JLog::DEBUG, 'ats.emails');

			$categoryID = $this->getCategoryIDFromEmail($email['toemail']);

			// If no category was found, use the default category
			if (empty($categoryID))
			{
				JLog::add(sprintf("No specific category corresponds to this email address; falling back to default category %d", $defaultCategory), JLog::DEBUG, 'ats.emails');

				$categoryID = $defaultCategory;
			}
			else
			{
				JLog::add(sprintf("Got category ID %d", $categoryID), JLog::DEBUG, 'ats.emails');
			}
		}
		else
		{
			$categoryID = $ticket->catid;

			JLog::add(sprintf("Got category ID %d from existing ticket #%u", $categoryID, $ticketID), JLog::DEBUG, 'ats.emails');
		}

		// Initialise
		/** @var User $user */
		$user = null;
		$from = $email->fromEmail;

		// Perform some generic checks. More specific ones are performed later
		if (!$this->passedBasicChecks($email, $categoryID, $ticket, $from, $user))
		{
			JLog::add("Generic checks failed for this email. Skipping this email.", JLog::ERROR, 'ats.emails');

			return false;
		}

		// Get the ACL privileges for the user
		$perms = Permissions::getAclPrivileges($categoryID, $user->id);

		/**
		 * The user account may have been found using an add-on email. We have to normalize this to the Joomla! user
		 * account email. This is the only way we can email ALL of the addresses (canonical and add-on) of the user.
		 */
		if ($user->id)
		{
			$from = $user->email;
		}

		$post      = [];
		$newTicket = !$ticket->getId();

		if ($newTicket)
		{
			JLog::add('Posting a new ticket', JLog::DEBUG, 'ats.emails');

			/**
			 * We have to decode the email subject, otherwise we end up with a wrong title when non UTF-8 chars are used
			 * (for example German umlauts)
			 */
			$title = $email['subject'];

			// TODO Public or private needs to take into account the category preferences as well
			$data = [
				'id'         => 0,
				'catid'      => $categoryID,
				'status'     => 'O',
				'title'      => $title,
				'public'     => $perms['ats.private'] ? 0 : 1,
				'origin'     => 'email',
				'enabled'    => 1,
				'created_by' => $user->id,
			];

			TicketTable::$overrideuser = true;

			try
			{
				if (!$ticket->save($data))
				{
					throw new RuntimeException($ticket->getError());
				}
			}
			catch (Exception $exception)
			{
				JLog::add("Ticket creation failed. Raw error message: " . $ticket->getError(), JLog::ERROR, 'ats.emails');

				JLog::add("\tException: " . $exception->getMessage(), JLog::ERROR, 'ats.emails');
				JLog::add("\tLine     : " . $exception->getFile() . '::L' . $exception->getLine(), JLog::ERROR, 'ats.emails');
				JLog::add("\tTraceback: " . $exception->getTraceAsString(), JLog::ERROR, 'ats.emails');

				// Send an email informing the user that ticket creation failed
				$this->sendErrorEmail('mailgateway_newticketfailed', $from, [
					'name'     => $user->name,
					'username' => $user->username,
					'user_id'  => $user->id,
				]);

				return false;
			}
			finally
			{
				TicketTable::$overrideuser = false;
			}
		}
		else
		{
			JLog::add("Ready to post a reply", JLog::DEBUG, 'ats.emails');
			$myTicket = $user->id == $ticket->created_by;

			$update = [
				'ticket_id'   => $ticket->id,
				'status'      => 'O',
				'modified_by' => $user->id,
				'modified_on' => gmdate($this->getDatabase()->getDateFormat()),
			];

			// Make sure the user is replying to his own ticket or he's an admin
			if ($myTicket)
			{
				JLog::add("Posting a reply from the ticket owner", JLog::DEBUG, 'ats.emails');

				// User posted: set ticket status to 'O'
				$update['status'] = 'O';
			}
			else
			{
				JLog::add("Posting a reply from a category manager", JLog::DEBUG, 'ats.emails');

				// Admin posted: set ticket status to 'P'
				$update['status'] = 'P';

				// If the user is a manager and there isn't an assigned user, let's assign him to the ticket
				if (!$ticket->assigned_to)
				{
					$update['assigned_to'] = $user->id;
				}
			}

			TicketTable::$noPermissionsCheck = true;

			try
			{
				if (!$ticket->save($update))
				{
					throw new RuntimeException($ticket->getError());
				}
			}
			catch (Exception $exception)
			{
				JLog::add("An error occurred updating the ticket status. Raw error: " . $ticket->getError(), JLog::ERROR, 'ats.emails');

				JLog::add("\tException: " . $exception->getMessage(), JLog::ERROR, 'ats.emails');
				JLog::add("\tLine     : " . $exception->getFile() . '::L' . $exception->getLine(), JLog::ERROR, 'ats.emails');
				JLog::add("\tTraceback: " . $exception->getTraceAsString(), JLog::ERROR, 'ats.emails');

				// Send an email informing the user that ticket creation failed
				$this->sendErrorEmail('mailgateway_newticketfailed', $from, [
					'name'     => $user->name,
					'username' => $user->username,
					'user_id'  => $user->id,
				]);

				return false;
			}
			finally
			{
				TicketTable::$noPermissionsCheck = false;
			}
		}

		$emailmessage = $this->extractMessage($email);
		$emailmessage = Filter::filterText($emailmessage, $user);

		// Add a post
		$post['content_html']  = $emailmessage;
		$post['enabled']       = 1;
		$post['origin']        = 'email';
		$post['created_by']    = $user->id;
		$post['email_uid']     = $email['uid'];
		$post['ticket_id']     = $ticket->getId();
		$post['attachment_id'] = '';

		if ($perms['ats.attachment'])
		{
			try
			{
				$attachments           = $this->extractAttachments($email, $user->id);
				$post['attachment_id'] = implode(',', $attachments);
			}
			catch (Exception $e)
			{
				/**
				 * This attachment failed to be uploaded, e.g. type mismatch. No problem; we are going to ignore it.
				 * If we don't we end up with a ticket that has no attachment *and* no body we're good.
				 */
			}
		}

		$hasAttachments = !empty($attachments);

		/** @var PostTable $postTable */
		$postTable = $this->getMVCFactory()->createTable('Post', 'Administrator');
		/** @var AttachmentTable $attachmentTable */
		$attachmentTable = $this->getMVCFactory()->createTable('Attachment', 'Administrator');

		try
		{
			$postTable->save($post);
		}
		catch (Exception $e)
		{
			// First, delete the attachments.
			if ($hasAttachments)
			{
				foreach ($attachments as $attachment_id)
				{
					try
					{
						$attachmentTable->delete($attachment_id);
					}
					catch (Exception $e)
					{
						// Don't die if we can't delete an orphan attachment.
					}
				}

			}

			// If an error occurred after saving the post, delete the post.
			if ($postTable->getId())
			{
				$postTable->delete();
			}

			// Finally, delete the ticket itself.
			if ($newTicket)
			{
				$ticket->delete();
			}


			JLog::add("Post creation failed. Raw error message: " . $e->getMessage(), JLog::ERROR, 'ats.emails');
			JLog::add("\tTraceback: " . $e->getTraceAsString(), JLog::ERROR, 'ats.emails');

			// Send an email informing the user that ticket creation failed
			$this->sendErrorEmail('mailgateway_newpostfailed', $from, [
				'name'     => $user->name,
				'username' => $user->username,
				'user_id'  => $user->id,
			]);

			return false;
		}

		// Update post reference inside each attachment record
		if ($hasAttachments)
		{
			foreach ($attachments as $attachment_id)
			{
				if ($attachmentTable->load($attachment_id))
				{
					$attachmentTable->post_id = $postTable->getId();
					$attachmentTable->store();
				}
			}
		}

		$this->postNotifiable($postTable, true);

		return true;
	}

	/**
	 * Sends an error email response to a client
	 *
	 * @param   string  $key        Email key to load
	 * @param   string  $toEmail    Email address to send the email to
	 * @param   array   $extraVars  Any additional variables used during template parsing
	 *
	 * @return  bool
	 *
	 * @since   3.2.0
	 */
	private function sendErrorEmail(string $key, string $toEmail, array $extraVars = []): bool
	{
		// Get the user, if we have a user ID. Otherwise get a guest user.
		$userId      = $extraVars['user_id'] ?? null;
		$user        = !empty($userId) ? Permissions::getUser($userId) : new User();
		$addonEmails = AddonEmails::getAddonEmails($user->id);

		// In case we have no matching user we still need to send the email. Pretend it's not a guest.
		if ($user->guest)
		{
			$user->name      = $toEmail;
			$user->guest     = 0;
			$user->block     = 0;
			$user->sendEmail = 1;
		}

		// Get a new mailer
		$mailer = clone Factory::getMailer();

		// Carbon copy any add–on email addresses
		foreach ($addonEmails as $email => $name)
		{
			$mailer->addCc($email, $name);
		}

		$app     = Factory::getApplication();
		$cParams = ComponentHelper::getParams('com_ats');

		$extraVars = array_merge([
			'SITENAME' => $app->get('sitename'),
			'SITEURL'  => $app->isClient('cli') ? $cParams->get('siteurl', 'https://www.example.com') : Uri::base(false),
			'NAME'     => $user->name ?: $user->email,
		], $extraVars);

		// Send the email
		return TemplateEmails::sendMail($key, $extraVars, $user, null, false, $mailer);
	}
}